# MIT License

# Copyright (c) Meta Platforms, Inc. and affiliates.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Copyright (c) [2023] [Meta Platforms, Inc. and affiliates.] 
# Copyright (c) [2025] [Ziyue Jiang] 
# SPDX-License-Identifier: MIT
# This file has been modified by Ziyue Jiang on 2025/03/19
# Original file was released under MIT, with the full license text # available at https://github.com/facebookresearch/encodec/blob/gh-pages/LICENSE.
# This modified file is released under the same license.

"""LSTM layers module."""
from torch import nn


class SLSTM(nn.Module):
    """
    LSTM without worrying about the hidden state, nor the layout of the data.
    Expects input as convolutional layout.
    """
    def __init__(self, dimension: int, num_layers: int = 2, skip: bool = True):
        super().__init__()
        self.skip = skip
        self.lstm = nn.LSTM(dimension, dimension, num_layers)

    # 修改transpose顺序
    def forward(self, x):
        x1 = x.permute(2, 0, 1)
        y, _ = self.lstm(x1)
        y = y.permute(1, 2, 0)
        if self.skip:
            y = y + x
        return y



import torch
from funasr_detach.optimizers.fairseq_adam import FairseqAdam
from funasr_detach.optimizers.sgd import SGD

optim_classes = dict(
    adam=torch.optim.Adam,
    fairseq_adam=FairseqAdam,
    adamw=torch.optim.AdamW,
    sgd=SGD,
    adadelta=torch.optim.Adadelta,
    adagrad=torch.optim.Adagrad,
    adamax=torch.optim.Adamax,
    asgd=torch.optim.ASGD,
    lbfgs=torch.optim.LBFGS,
    rmsprop=torch.optim.RMSprop,
    rprop=torch.optim.Rprop,
)



# Copyright (c) 2023-2024 DeepSeek.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import torch
from transformers import AutoModelForCausalLM

from janus.models import MultiModalityCausalLM, VLChatProcessor
import numpy as np
import os
import PIL.Image

# specify the path to the model
model_path = "deepseek-ai/Janus-1.3B"
vl_chat_processor: VLChatProcessor = VLChatProcessor.from_pretrained(model_path)
tokenizer = vl_chat_processor.tokenizer

vl_gpt: MultiModalityCausalLM = AutoModelForCausalLM.from_pretrained(
    model_path, trust_remote_code=True
)
vl_gpt = vl_gpt.to(torch.bfloat16).cuda().eval()

conversation = [
    {
        "role": "User",
        "content": "A close-up high-contrast photo of Sydney Opera House sitting next to Eiffel tower, under a blue night sky of roiling energy, exploding yellow stars, and radiating swirls of blue.",
    },
    {"role": "Assistant", "content": ""},
]

sft_format = vl_chat_processor.apply_sft_template_for_multi_turn_prompts(
    conversations=conversation,
    sft_format=vl_chat_processor.sft_format,
    system_prompt="",
)
prompt = sft_format + vl_chat_processor.image_start_tag


@torch.inference_mode()
def generate(
    mmgpt: MultiModalityCausalLM,
    vl_chat_processor: VLChatProcessor,
    prompt: str,
    temperature: float = 1,
    parallel_size: int = 16,
    cfg_weight: float = 5,
    image_token_num_per_image: int = 576,
    img_size: int = 384,
    patch_size: int = 16,
):
    input_ids = vl_chat_processor.tokenizer.encode(prompt)
    input_ids = torch.LongTensor(input_ids)

    tokens = torch.zeros((parallel_size*2, len(input_ids)), dtype=torch.int).cuda()
    for i in range(parallel_size*2):
        tokens[i, :] = input_ids
        if i % 2 != 0:
            tokens[i, 1:-1] = vl_chat_processor.pad_id

    inputs_embeds = mmgpt.language_model.get_input_embeddings()(tokens)

    generated_tokens = torch.zeros((parallel_size, image_token_num_per_image), dtype=torch.int).cuda()

    for i in range(image_token_num_per_image):
        outputs = mmgpt.language_model.model(inputs_embeds=inputs_embeds, use_cache=True, past_key_values=outputs.past_key_values if i != 0 else None)
        hidden_states = outputs.last_hidden_state
        
        logits = mmgpt.gen_head(hidden_states[:, -1, :])
        logit_cond = logits[0::2, :]
        logit_uncond = logits[1::2, :]
        
        logits = logit_uncond + cfg_weight * (logit_cond-logit_uncond)
        probs = torch.softmax(logits / temperature, dim=-1)

        next_token = torch.multinomial(probs, num_samples=1)
        generated_tokens[:, i] = next_token.squeeze(dim=-1)

        next_token = torch.cat([next_token.unsqueeze(dim=1), next_token.unsqueeze(dim=1)], dim=1).view(-1)
        img_embeds = mmgpt.prepare_gen_img_embeds(next_token)
        inputs_embeds = img_embeds.unsqueeze(dim=1)


    dec = mmgpt.gen_vision_model.decode_code(generated_tokens.to(dtype=torch.int), shape=[parallel_size, 8, img_size//patch_size, img_size//patch_size])
    dec = dec.to(torch.float32).cpu().numpy().transpose(0, 2, 3, 1)

    dec = np.clip((dec + 1) / 2 * 255, 0, 255)

    visual_img = np.zeros((parallel_size, img_size, img_size, 3), dtype=np.uint8)
    visual_img[:, :, :] = dec

    os.makedirs('generated_samples', exist_ok=True)
    for i in range(parallel_size):
        save_path = os.path.join('generated_samples', "img_{}.jpg".format(i))
        PIL.Image.fromarray(visual_img[i]).save(save_path)


generate(
    vl_gpt,
    vl_chat_processor,
    prompt,
)


import logging
import os
from pathlib import Path
import sys
import time
from typing import AsyncGenerator

import numpy as np
import torch
import torchaudio

try:
    cur_dir = os.path.dirname(__file__)
    if bool(os.getenv("ACHATBOT_PKG", "")):
        sys.path.insert(1, os.path.join(cur_dir, "../../../Zonos"))
    else:
        sys.path.insert(1, os.path.join(cur_dir, "../../../../deps/Zonos"))
    from deps.Zonos.zonos.model import Zonos
    from deps.Zonos.zonos.conditioning import make_cond_dict
except ModuleNotFoundError as e:
    logging.error(
        "In order to use zonos-tts use transformer, you need to `apt install -y espeak-ng`,  `pip install achatbot[tts_zonos]`."
        + "if use transformer + mabama2, need `pip install achatbot[tts_zonos_hybrid]`"
    )
    raise Exception(f"Missing module: {e}")

from src.common.utils.audio_utils import AUDIO_EXTENSIONS
from src.common.random import set_all_random_seed
from src.common.types import PYAUDIO_PAFLOAT32, PYAUDIO_PAINT16
from src.common.utils.helper import file_md5_hash, get_device, print_model_params
from src.common.interface import ITts
from src.common.session import Session
from src.types.speech.tts.zonos import ZonosTTSArgs
from .base import BaseTTS


class ZonosSpeechTTS(BaseTTS, ITts):
    TAG = "tts_zonos"

    @classmethod
    def get_args(cls, **kwargs) -> dict:
        return {**ZonosTTSArgs().__dict__, **kwargs}

    def __init__(self, **args) -> None:
        self.voices = {}
        self.args = ZonosTTSArgs(**args)
        self.args.device = self.args.device or get_device()
        logging.debug(f"args:{self.args}")
        repo_id = "/".join(self.args.lm_checkpoint_dir.split("/")[-2:])
        self.model = Zonos.from_pretrained(
            repo_id,
            device=self.args.device,
            local_dir=self.args.lm_checkpoint_dir,
            dac_model_path=self.args.dac_model_dir,
        )
        self.model.requires_grad_(False).eval()

        self.speaker: torch.Tensor = None
        if self.args.ref_audio_file_path and os.path.exists(self.args.ref_audio_file_path):
            wav, sr = torchaudio.load(self.args.ref_audio_file_path)
            self.speaker = self.model.make_speaker_embedding(
                wav, sr, local_dir=self.args.speaker_embedding_model_dir
            )

        self.warm_up()

    def warm_up(self):
        if not self.args.warm_up_text:
            logging.warning(f"No warm_up_text to Warm Up")
            return

        logging.info(f"Warming up {self.__class__.__name__} device: {self.model.device}")
        cond_dict = make_cond_dict(
            text=self.args.warm_up_text, speaker=self.speaker, language=self.args.language
        )
        conditioning = self.model.prepare_conditioning(cond_dict)
        stream_generator = self.model.stream(
            prefix_conditioning=conditioning,
            audio_prefix_codes=None,  # no audio prefix
            chunk_size=self.args.chunk_size,
        )

        if "cuda" in str(self.model.device):
            start_event = torch.cuda.Event(enable_timing=True)
            end_event = torch.cuda.Event(enable_timing=True)
            torch.cuda.synchronize()
            start_event.record()

        for step in range(self.args.warmup_steps):
            times = []
            start_time = time.perf_counter()
            for _ in stream_generator:
                times.append(time.perf_counter() - start_time)
                start_time = time.perf_counter()
            logging.info(f"step {step} warnup TTFT(chunk) time: {times[0]} s")

        if "cuda" in str(self.model.device):
            end_event.record()
            torch.cuda.synchronize()
            logging.info(
                f"{self.__class__.__name__}:  warmed up! time: {start_event.elapsed_time(end_event) * 1e-3:.3f} s"
            )
        logging.info(f"End Warm Up")

    def set_voice(self, ref_audio_path: str):
        """
        - ref_audio_path is audio path, make speaker embedding
        """
        if os.path.exists(ref_audio_path) is False:
            raise FileNotFoundError(f"reference_audio_path: {ref_audio_path}")

        md5_hash = file_md5_hash(ref_audio_path)
        if md5_hash in self.voices:
            logging.info(f"{ref_audio_path} had set speaker embedding")
            return

        wav, sr = torchaudio.load(ref_audio_path)
        self.speaker = self.model.make_speaker_embedding(
            wav, sr, local_dir=self.args.speaker_embedding_model_dir
        )
        self.voices[md5_hash] = self.speaker

    def get_voices(self) -> list:
        return list(self.voices.keys())

    def get_stream_info(self) -> dict:
        return {
            # "format": PYAUDIO_PAINT16,
            "format": PYAUDIO_PAFLOAT32,
            "channels": 1,
            # "rate": self.gan_model.spec_transform.sample_rate,
            # https://huggingface.co/descript/dac_44khz/blob/main/config.json
            "rate": self.model.autoencoder.sampling_rate,  # 44_100
            "sample_width": 2,
            # "np_dtype": np.int16,
            "np_dtype": np.float32,
        }

    async def _inference(
        self, session: Session, text: str, **kwargs
    ) -> AsyncGenerator[bytes, None]:
        # logging.debug(f"session:{session}, text:{text}")
        if "cuda" in str(self.model.device):
            torch.cuda.empty_cache()
        seed = kwargs.get("seed", self.args.seed)

        set_all_random_seed(seed)

        # default yield every 40 generated tokens,
        # less for faster streaming but lower quality
        chunk_size = kwargs.get("chunk_size", self.args.chunk_size)
        language = kwargs.get("language", self.args.language)
        emotion = kwargs.get("emotion", self.args.emotion)
        fmax = kwargs.get("fmax", self.args.fmax)
        pitch_std = kwargs.get("pitch_std", self.args.pitch_std)
        speaking_rate = kwargs.get("speaking_rate", self.args.speaking_rate)
        vqscore_8 = kwargs.get("vqscore_8", self.args.vqscore_8)
        ctc_loss = kwargs.get("ctc_loss", self.args.ctc_loss)
        dnsmos_ovrl = kwargs.get("dnsmos_ovrl", self.args.dnsmos_ovrl)

        # Create the conditioning dictionary (using text, speaker embedding, language, etc.).
        cond_dict = make_cond_dict(
            text=text,
            speaker=self.speaker,
            language=language,
            emotion=emotion,
            fmax=fmax,
            pitch_std=pitch_std,
            speaking_rate=speaking_rate,
            vqscore_8=vqscore_8,
            ctc_loss=ctc_loss,
            dnsmos_ovrl=dnsmos_ovrl,
        )
        conditioning = self.model.prepare_conditioning(cond_dict)

        logging.debug("Starting streaming generation...")
        stream_generator = self.model.stream(
            prefix_conditioning=conditioning,
            audio_prefix_codes=None,  # no audio prefix in this test
            chunk_size=chunk_size,
        )

        for sr_out, codes_chunk in stream_generator:
            logging.debug(f"Received codes chunk of shape: {codes_chunk.shape}, rate: {sr_out}")
            # [1,1,T]->[T]
            tensor_audio_chunk = self.model.autoencoder.decode(codes_chunk)[0][0]
            # Convert the audio to numpy(float32)
            nparr_audio_chunk = tensor_audio_chunk.detach().cpu().numpy()

            yield nparr_audio_chunk.tobytes()



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

"""Convolutional layers wrappers and utilities."""

import math
import typing as tp
import warnings

import torch
from torch import nn
from torch.nn import functional as F
from torch.nn.utils import spectral_norm, weight_norm

from .norm import ConvLayerNorm


CONV_NORMALIZATIONS = frozenset(['none', 'weight_norm', 'spectral_norm',
                                 'time_layer_norm', 'layer_norm', 'time_group_norm'])


def apply_parametrization_norm(module: nn.Module, norm: str = 'none') -> nn.Module:
    assert norm in CONV_NORMALIZATIONS
    if norm == 'weight_norm':
        return weight_norm(module)
    elif norm == 'spectral_norm':
        return spectral_norm(module)
    else:
        # We already check was in CONV_NORMALIZATION, so any other choice
        # doesn't need reparametrization.
        return module


def get_norm_module(module: nn.Module, causal: bool = False, norm: str = 'none', **norm_kwargs) -> nn.Module:
    """Return the proper normalization module. If causal is True, this will ensure the returned
    module is causal, or return an error if the normalization doesn't support causal evaluation.
    """
    assert norm in CONV_NORMALIZATIONS
    if norm == 'layer_norm':
        assert isinstance(module, nn.modules.conv._ConvNd)
        return ConvLayerNorm(module.out_channels, **norm_kwargs)
    elif norm == 'time_group_norm':
        if causal:
            raise ValueError("GroupNorm doesn't support causal evaluation.")
        assert isinstance(module, nn.modules.conv._ConvNd)
        return nn.GroupNorm(1, module.out_channels, **norm_kwargs)
    else:
        return nn.Identity()


def get_extra_padding_for_conv1d(x: torch.Tensor, kernel_size: int, stride: int,
                                 padding_total: int = 0) -> int:
    """See `pad_for_conv1d`.
    """
    length = x.shape[-1]
    n_frames = (length - kernel_size + padding_total) / stride + 1
    ideal_length = (math.ceil(n_frames) - 1) * stride + (kernel_size - padding_total)
    return ideal_length - length


def pad_for_conv1d(x: torch.Tensor, kernel_size: int, stride: int, padding_total: int = 0):
    """Pad for a convolution to make sure that the last window is full.
    Extra padding is added at the end. This is required to ensure that we can rebuild
    an output of the same length, as otherwise, even with padding, some time steps
    might get removed.
    For instance, with total padding = 4, kernel size = 4, stride = 2:
        0 0 1 2 3 4 5 0 0   # (0s are padding)
        1   2   3           # (output frames of a convolution, last 0 is never used)
        0 0 1 2 3 4 5 0     # (output of tr. conv., but pos. 5 is going to get removed as padding)
            1 2 3 4         # once you removed padding, we are missing one time step !
    """
    extra_padding = get_extra_padding_for_conv1d(x, kernel_size, stride, padding_total)
    return F.pad(x, (0, extra_padding))


def pad1d(x: torch.Tensor, paddings: tp.Tuple[int, int], mode: str = 'zero', value: float = 0.):
    """Tiny wrapper around F.pad, just to allow for reflect padding on small input.
    If this is the case, we insert extra 0 padding to the right before the reflection happen.
    """
    length = x.shape[-1]
    padding_left, padding_right = paddings
    assert padding_left >= 0 and padding_right >= 0, (padding_left, padding_right)
    if mode == 'reflect':
        max_pad = max(padding_left, padding_right)
        extra_pad = 0
        if length <= max_pad:
            extra_pad = max_pad - length + 1
            x = F.pad(x, (0, extra_pad))
        padded = F.pad(x, paddings, mode, value)
        end = padded.shape[-1] - extra_pad
        return padded[..., :end]
    else:
        return F.pad(x, paddings, mode, value)


def unpad1d(x: torch.Tensor, paddings: tp.Tuple[int, int]):
    """Remove padding from x, handling properly zero padding. Only for 1d!"""
    padding_left, padding_right = paddings
    assert padding_left >= 0 and padding_right >= 0, (padding_left, padding_right)
    assert (padding_left + padding_right) <= x.shape[-1]
    end = x.shape[-1] - padding_right
    return x[..., padding_left: end]


class NormConv1d(nn.Module):
    """Wrapper around Conv1d and normalization applied to this conv
    to provide a uniform interface across normalization approaches.
    """
    def __init__(self, *args, causal: bool = False, norm: str = 'none',
                 norm_kwargs: tp.Dict[str, tp.Any] = {}, **kwargs):
        super().__init__()
        self.conv = apply_parametrization_norm(nn.Conv1d(*args, **kwargs), norm)
        self.norm = get_norm_module(self.conv, causal, norm, **norm_kwargs)
        self.norm_type = norm

    def forward(self, x):
        x = self.conv(x)
        x = self.norm(x)
        return x


class NormConv2d(nn.Module):
    """Wrapper around Conv2d and normalization applied to this conv
    to provide a uniform interface across normalization approaches.
    """
    def __init__(self, *args, norm: str = 'none',
                 norm_kwargs: tp.Dict[str, tp.Any] = {}, **kwargs):
        super().__init__()
        self.conv = apply_parametrization_norm(nn.Conv2d(*args, **kwargs), norm)
        self.norm = get_norm_module(self.conv, causal=False, norm=norm, **norm_kwargs)
        self.norm_type = norm

    def forward(self, x):
        x = self.conv(x)
        x = self.norm(x)
        return x


class NormConvTranspose1d(nn.Module):
    """Wrapper around ConvTranspose1d and normalization applied to this conv
    to provide a uniform interface across normalization approaches.
    """
    def __init__(self, *args, causal: bool = False, norm: str = 'none',
                 norm_kwargs: tp.Dict[str, tp.Any] = {}, **kwargs):
        super().__init__()
        self.convtr = apply_parametrization_norm(nn.ConvTranspose1d(*args, **kwargs), norm)
        self.norm = get_norm_module(self.convtr, causal, norm, **norm_kwargs)
        self.norm_type = norm

    def forward(self, x):
        x = self.convtr(x)
        x = self.norm(x)
        return x


class NormConvTranspose2d(nn.Module):
    """Wrapper around ConvTranspose2d and normalization applied to this conv
    to provide a uniform interface across normalization approaches.
    """
    def __init__(self, *args, norm: str = 'none',
                 norm_kwargs: tp.Dict[str, tp.Any] = {}, **kwargs):
        super().__init__()
        self.convtr = apply_parametrization_norm(nn.ConvTranspose2d(*args, **kwargs), norm)
        self.norm = get_norm_module(self.convtr, causal=False, norm=norm, **norm_kwargs)

    def forward(self, x):
        x = self.convtr(x)
        x = self.norm(x)
        return x


class SConv1d(nn.Module):
    """Conv1d with some builtin handling of asymmetric or causal padding
    and normalization.
    """
    def __init__(self, in_channels: int, out_channels: int,
                 kernel_size: int, stride: int = 1, dilation: int = 1,
                 groups: int = 1, bias: bool = True, causal: bool = False,
                 norm: str = 'none', norm_kwargs: tp.Dict[str, tp.Any] = {},
                 pad_mode: str = 'reflect'):
        super().__init__()
        # warn user on unusual setup between dilation and stride
        if stride > 1 and dilation > 1:
            warnings.warn('SConv1d has been initialized with stride > 1 and dilation > 1'
                          f' (kernel_size={kernel_size} stride={stride}, dilation={dilation}).')
        self.conv = NormConv1d(in_channels, out_channels, kernel_size, stride,
                               dilation=dilation, groups=groups, bias=bias, causal=causal,
                               norm=norm, norm_kwargs=norm_kwargs)
        self.causal = causal
        self.pad_mode = pad_mode

    def forward(self, x):
        B, C, T = x.shape
        kernel_size = self.conv.conv.kernel_size[0]
        stride = self.conv.conv.stride[0]
        dilation = self.conv.conv.dilation[0]
        kernel_size = (kernel_size - 1) * dilation + 1  # effective kernel size with dilations
        padding_total = kernel_size - stride
        extra_padding = get_extra_padding_for_conv1d(x, kernel_size, stride, padding_total)
        if self.causal:
            # Left padding for causal
            x = pad1d(x, (padding_total, extra_padding), mode=self.pad_mode)
        else:
            # Asymmetric padding required for odd strides
            padding_right = padding_total // 2
            padding_left = padding_total - padding_right
            x = pad1d(x, (padding_left, padding_right + extra_padding), mode=self.pad_mode)
        return self.conv(x)


class SConvTranspose1d(nn.Module):
    """ConvTranspose1d with some builtin handling of asymmetric or causal padding
    and normalization.
    """
    def __init__(self, in_channels: int, out_channels: int,
                 kernel_size: int, stride: int = 1, causal: bool = False,
                 norm: str = 'none', trim_right_ratio: float = 1.,
                 norm_kwargs: tp.Dict[str, tp.Any] = {}):
        super().__init__()
        self.convtr = NormConvTranspose1d(in_channels, out_channels, kernel_size, stride,
                                          causal=causal, norm=norm, norm_kwargs=norm_kwargs)
        self.causal = causal
        self.trim_right_ratio = trim_right_ratio
        assert self.causal or self.trim_right_ratio == 1., \
            "`trim_right_ratio` != 1.0 only makes sense for causal convolutions"
        assert self.trim_right_ratio >= 0. and self.trim_right_ratio <= 1.

    def forward(self, x):
        kernel_size = self.convtr.convtr.kernel_size[0]
        stride = self.convtr.convtr.stride[0]
        padding_total = kernel_size - stride

        y = self.convtr(x)

        # We will only trim fixed padding. Extra padding from `pad_for_conv1d` would be
        # removed at the very end, when keeping only the right length for the output,
        # as removing it here would require also passing the length at the matching layer
        # in the encoder.
        if self.causal:
            # Trim the padding on the right according to the specified ratio
            # if trim_right_ratio = 1.0, trim everything from right
            padding_right = math.ceil(padding_total * self.trim_right_ratio)
            padding_left = padding_total - padding_right
            y = unpad1d(y, (padding_left, padding_right))
        else:
            # Asymmetric padding required for odd strides
            padding_right = padding_total // 2
            padding_left = padding_total - padding_right
            y = unpad1d(y, (padding_left, padding_right))
        return y



# Copyright 2025 ByteDance and/or its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import torch.nn.functional as F

def map_phone_to_tokendict(item, pad_bos_eos=True):
    # Merge Chinese phone and tone (Original dict ends at 173, i.e., ph_dict_size=173). 146~173 is punctuations.
    phone = item['txt_token'].clone()
    merged_phone = item['txt_token'].clone()
    tone_tmp = item['tone'].clone()
    # In tone_dict, tone_1 is 4, tone_2 is 11, tone_3 is 12, tone_4 is 13, tone_5 is 14, tone_6 is 15
    tone_tmp[tone_tmp==4] = 1
    tone_tmp[tone_tmp==11] = 2
    tone_tmp[tone_tmp==12] = 3
    tone_tmp[tone_tmp==13] = 4
    tone_tmp[tone_tmp==14] = 5
    tone_tmp[tone_tmp==15] = 6
    # Chinese phones lie in 3~100 in the phone_dict, we map them to 200~788
    ch_phone_idx = (phone >= 3) & (phone <= 100)
    merged_phone[ch_phone_idx] = (merged_phone[ch_phone_idx] - 3) * 6 + 200 + tone_tmp[ch_phone_idx]

    if pad_bos_eos:
        merged_phone = F.pad(merged_phone, (1, 0), mode='constant', value=798)
        merged_phone = F.pad(merged_phone, (0, 1), mode='constant', value=799)
    return merged_phone
    
def split_ph_timestamp(ph_timestamp):
    ''' Input: ph_timestamp, shape [T] '''

    # Map the timestamp of each phone back to its original frame-level lengths
    ph_timestamp[ph_timestamp >= 800] -= 800

    ph_list = []
    tone_list = []
    dur_list = []
    cur_timestamp = 0
    for idx, item in enumerate(ph_timestamp):
        if idx % 2 == 0:
            # Map Chinese phones back to its original phone_dict
            if (200 <= item <= 788):
                ph = (item - 200 - 1) // 6 + 3
                tone = (item - 200 - 1) % 6 + 1
                if tone == 1:
                    tone = 4
                else:
                    tone = tone + 9
            # Set English tone to '3'
            else:
                ph = item
                tone = 3
            ph_list.append(ph)
            tone_list.append(tone)
        else:
            dur_list.append((item - cur_timestamp))
            cur_timestamp = item
    assert len(ph_list) == len(dur_list), f"{len(ph_list)}, {len(dur_list)}"
    ph_seq, tone_seq, dur_seq = torch.LongTensor(ph_list), torch.LongTensor(tone_list), torch.LongTensor(dur_list)
    return ph_seq, tone_seq, dur_seq, ph_timestamp[-1]
    
def split_ph(ph_seq):
    ''' Input: ph_timestamp, shape [T] '''
    ph_list = []
    tone_list = []
    for idx, item in enumerate(ph_seq):
        # Map Chinese phones back to its original phone_dict
        if (200 <= item <= 788):
            ph = (item - 200 - 1) // 6 + 3
            tone = (item - 200 - 1) % 6 + 1
            if tone == 1:
                tone = 4
            else:
                tone = tone + 9
        # Set English tone to '3'
        else:
            ph = item
            tone = 3
        ph_list.append(ph)
        tone_list.append(tone)
        
    assert len(ph_list) == len(tone_list)
    ph_seq, tone_seq = torch.LongTensor(ph_list), torch.LongTensor(tone_list)
    return ph_seq, tone_seq


import torch
import torch.nn as nn
import torch.nn.functional as F


def find_multiple(n: int, k: int) -> int:
    if k == 0 or n % k == 0:
        return n
    return n + k - (n % k)


def pad_weight_(w: nn.Embedding | nn.Linear, multiple: int):
    """Pad the weight of an embedding or linear layer to a multiple of `multiple`."""
    if isinstance(w, nn.Embedding):
        # Pad input dim
        if w.weight.shape[1] % multiple == 0:
            return
        w.weight.data = F.pad(w.weight.data, (0, 0, 0, w.weight.shape[1] % multiple))
        w.num_embeddings, w.embedding_dim = w.weight.shape
    elif isinstance(w, nn.Linear):
        # Pad output dim
        if w.weight.shape[0] % multiple == 0:
            return
        w.weight.data = F.pad(w.weight.data, (0, 0, 0, w.weight.shape[0] % multiple))
        w.out_features, w.in_features = w.weight.shape
    else:
        raise ValueError(f"Unsupported weight type: {type(w)}")


def get_device() -> torch.device:
    if torch.cuda.is_available():
        return torch.device(torch.cuda.current_device())
    # MPS breaks for whatever reason. Uncomment when it's working.
    # if torch.mps.is_available():
    #     return torch.device("mps")
    return torch.device("cpu")


DEFAULT_DEVICE = get_device()



import modal
import os

app = modal.App("download_vllm_bench_dataset")

# Define the dependencies for the function using a Modal Image.
download_image = (
    modal.Image.debian_slim(python_version="3.11")
    .apt_install("git", "curl", "wget", "unzip")  # Install curl for downloading files
    .pip_install("huggingface_hub")
)

BENCH_DIR = "/data/bench"
bench_dir = modal.Volume.from_name("bench", create_if_missing=True)


@app.function(
    # gpu="T4",
    retries=2,
    cpu=2.0,
    image=download_image,
    # secrets=[modal.Secret.from_name("achatbot")],
    volumes={BENCH_DIR: bench_dir},
    timeout=1200,
    scaledown_window=1200,
)
def download_ShareGPT() -> str:
    """
    Downloads ShareGPT and saves them to the BENCH_DIR.

    Returns:
        A message indicating that the downloads are complete.
    """
    import subprocess
    import logging

    logging.basicConfig(level=logging.INFO)

    os.makedirs(BENCH_DIR, exist_ok=True)

    try:
        subprocess.run(
            [
                "wget",
                "https://huggingface.co/datasets/anon8231489123/ShareGPT_Vicuna_unfiltered/resolve/main/ShareGPT_V3_unfiltered_cleaned_split.json",
                "-O",
                "/data/bench/ShareGPT_V3_unfiltered_cleaned_split.json",
            ],
            capture_output=True,
            text=True,
            check=True,
        )
        logging.info(f"Download of ShareGPT complete.")

    except subprocess.CalledProcessError as e:
        logging.error(f"Error downloading ShareGPT: {e}")
        logging.error(f"curl stderr: {e.stderr}")

    return "All ShareGPT downloads complete."


@app.function(
    # gpu="T4",
    retries=2,
    cpu=2.0,
    image=download_image,
    volumes={BENCH_DIR: bench_dir},
    # timeout=1200,
    # scaledown_window=1200,
)
def download_ShareGPT4V() -> str:
    """
    Downloads ShareGPT4V and saves them to the BENCH_DIR.

    Returns:
        A message indicating that the downloads are complete.
    """
    import subprocess
    import logging

    logging.basicConfig(level=logging.INFO)

    os.makedirs(BENCH_DIR, exist_ok=True)
    os.makedirs(BENCH_DIR + "/coco", exist_ok=True)

    try:
        if os.path.exists("/data/bench/sharegpt4v_instruct_gpt4-vision_cap100k.json"):
            logging.info("File sharegpt4v_instruct_gpt4-vision_cap100k.json already exists.")
        else:
            subprocess.run(
                [
                    "wget",
                    "https://huggingface.co/datasets/Lin-Chen/ShareGPT4V/resolve/main/sharegpt4v_instruct_gpt4-vision_cap100k.json",
                    "-O",
                    "/data/bench/sharegpt4v_instruct_gpt4-vision_cap100k.json",
                ],
                capture_output=True,
                text=True,
                check=True,
            )
            logging.info(f"Download of ShareGPT complete.")
        if os.path.exists("/data/bench/coco/train2017.zip"):
            logging.info("File /data/bench/coco/train2017.zip already exists.")
        else:
            subprocess.run(
                [
                    "wget",
                    "http://images.cocodataset.org/zips/train2017.zip",
                    "-O",
                    "/data/bench/coco/train2017.zip",
                ],
                capture_output=True,
                text=True,
                check=True,
            )
        if os.path.exists("/data/bench/coco/train2017.zip.done"):
            logging.info("File /data/bench/coco/train2017.zip.done already exists.")
        else:
            logging.info("unzip /data/bench/coco/train2017.zip")
            subprocess.run(
                [
                    "unzip",
                    "/data/bench/coco/train2017.zip",
                    "-d",
                    "/data/bench/coco/",
                ],
                capture_output=True,
                text=True,
                check=True,
            )
            with open("/data/bench/coco/train2017.zip.done", "w"):
                logging.info("unzip /data/bench/coco/train2017.zip done")
                pass

        logging.info(f"Download of ShareGPT4V complete.")

    except subprocess.CalledProcessError as e:
        logging.error(f"Error downloading ShareGPT: {e}")
        logging.error(f"curl stderr: {e.stderr}")

    return "All ShareGPT downloads complete."


@app.function(
    # gpu="T4",
    retries=2,
    cpu=2.0,
    image=download_image,
    volumes={BENCH_DIR: bench_dir},
    timeout=1200,
    scaledown_window=1200,
)
def download_BurstGPT() -> str:
    """
    Downloads BurstGPT and saves them to the BENCH_DIR.

    Returns:
        A message indicating that the downloads are complete.
    """
    import subprocess
    import logging

    logging.basicConfig(level=logging.INFO)

    os.makedirs(BENCH_DIR, exist_ok=True)

    try:
        subprocess.run(
            [
                "wget",
                "https://github.com/HPMLL/BurstGPT/releases/download/v1.1/BurstGPT_without_fails_2.csv",
                "-O",
                "/data/bench/BurstGPT_without_fails_2.csv",
            ],
            capture_output=True,
            text=True,
            check=True,
        )
        logging.info(f"Download of BurstGPT complete.")

    except subprocess.CalledProcessError as e:
        logging.error(f"Error downloading BurstGPT: {e}")
        logging.error(f"curl stderr: {e.stderr}")

    return "All BurstGPT downloads complete."


"""
modal run src/llm/vllm/bench/download_dataset.py --dataset ShareGPT
modal run src/llm/vllm/bench/download_dataset.py --dataset ShareGPT4V
modal run src/llm/vllm/bench/download_dataset.py --dataset BurstGPT
"""


@app.local_entrypoint()
def main(dataset: str = "ShareGPT"):
    """
    Local entry point to trigger the asset download function.
    """
    if dataset == "ShareGPT":
        download_ShareGPT.remote()
    elif dataset == "ShareGPT4V":
        download_ShareGPT4V.remote()
    elif dataset == "BurstGPT":
        download_BurstGPT.remote()



import random
from dataclasses import dataclass
from itertools import chain
from pathlib import Path
from random import Random
from typing import Optional, Union

import numpy as np
import pyarrow.parquet as pq
import torch
import torch.nn.functional as F
from datasets.download.streaming_download_manager import xopen
from huggingface_hub import HfApi
from lightning import LightningDataModule
from torch.distributed import get_rank, get_world_size, is_initialized
from torch.utils.data import DataLoader, Dataset, IterableDataset, get_worker_info

from fish_speech.conversation import (
    CODEBOOK_PAD_TOKEN_ID,
    Conversation,
    Message,
    TextPart,
    VQPart,
)
from fish_speech.datasets.protos.text_data_pb2 import SampledData
from fish_speech.datasets.protos.text_data_stream import read_pb_stream
from fish_speech.text.clean import clean_text
from fish_speech.tokenizer import FishTokenizer
from fish_speech.utils import RankedLogger
from fish_speech.utils.braceexpand import braceexpand

log = RankedLogger(__name__, rank_zero_only=True)


def split_by_rank_worker(files):
    # We need to know the total number of devices
    # to split the data properly

    total_devices = 1
    if is_initialized():
        total_devices = get_world_size()

    worker_info = get_worker_info()
    if worker_info is not None:
        total_devices *= worker_info.num_workers

    if len(files) < total_devices:
        # Repeat the files N times to match the number of devices
        files = files * (total_devices // len(files) + 1)

    # DDP
    if is_initialized():
        files = files[get_rank() :: get_world_size()]

    # Split by worker
    if worker_info is not None:
        files = files[worker_info.id :: worker_info.num_workers]

    return files


class AutoTextSemanticInstructionIterableDataset(IterableDataset):
    """
    Auto Augment Dataset by Speaker

    1. Random concatenate multiple sentences from the same speaker to form a longer sentence
    2. Automatically normalize the text

    For interactive mode, we use the following format (multiple sequences):
    <s> [INST] [SPK: speaker] text [/INST] ... [INST] text [/INST] </s>

    For non-interactive mode, we use the following format (one long sequence):
    <s> [INST] text [/INST] ... </s>
    """

    def __init__(
        self,
        proto_files: list[str],
        seed: int = 42,
        interactive_prob: float = 0.5,
        max_length: int = 1024,
        tokenizer: FishTokenizer = None,
        use_speaker: bool | float = True,
        causal: bool = True,
        num_codebooks: Optional[int] = None,
        skip_text_prob: float = 0.0,
    ):
        """
        Args:
            proto_files: proto buf files if using local data
            seed: random seed
            interactive_prob: probability to use interactive mode
            max_length: max length of the text
            tokenizer: tokenizer
            use_speaker: include speaker information in the prompt
            causal: use causal sampling when using local data, disable will lead to random sampling
            num_codebooks: number of codebooks, if None, it will be automatically detected
            skip_text_prob: probability to skip the text (audio only), this only applies to interactive mode
        """

        super().__init__()

        assert 0 <= interactive_prob <= 1, "interactive_prob must be in [0, 1]"

        self.seed = seed
        self.max_length = max_length
        self.tokenizer = tokenizer
        self.interactive_prob = interactive_prob
        self.use_speaker = use_speaker
        self.proto_files = proto_files
        self.causal = causal
        self.num_codebooks = num_codebooks
        self.skip_text_prob = skip_text_prob

        self.groups = None

    def __iter__(self):
        while True:
            yield self.augment()

    def init_mock_data_server(self):
        if self.groups is not None:
            return

        # Expand the proto files
        expanded_proto_files = []
        for filename in self.proto_files:
            for i in braceexpand(filename):
                i = Path(i)
                if i.is_file():
                    expanded_proto_files.append(i)
                elif i.is_dir():
                    expanded_proto_files.extend(i.rglob("*.proto"))
                    expanded_proto_files.extend(i.rglob("*.protos"))
                else:
                    raise ValueError(f"{i} is not a file or directory")

        expanded_proto_files = sorted(expanded_proto_files)
        Random(self.seed).shuffle(expanded_proto_files)

        self.groups = []
        shard_proto_files = split_by_rank_worker(expanded_proto_files)
        log.info(
            f"Reading {len(shard_proto_files)} / {len(expanded_proto_files)} files"
        )

        count = 0
        for filename in shard_proto_files:
            with open(filename, "rb") as f:
                for text_data in read_pb_stream(f):
                    self.groups.append(text_data)
                    count += 1

        log.info(f"Read total {count} groups of data")

        # Shuffle the lines
        Random(self.seed).shuffle(self.groups)
        self.group_weights = [len(i.sentences) for i in self.groups]

    def sample_data(self):
        if self.groups is None:
            self.init_mock_data_server()

        # Shuffle unique lines, estimate that each sample is at least 20 tokens
        num_samples = self.max_length // 20

        # choice group based on their number of samples
        group = random.choices(self.groups, weights=self.group_weights, k=1)[0]

        if self.causal:
            # Sample in order
            if num_samples >= len(group.sentences):
                samples = group.sentences
            else:
                begin = random.randint(0, len(group.sentences) - num_samples)
                samples = group.sentences[begin : begin + num_samples]
        else:
            samples = random.choices(
                group.sentences, k=min(num_samples, len(group.sentences))
            )

        return SampledData(
            source=group.source,
            name=group.name,
            samples=samples,
        )

    def pack_sentences(
        self,
        sentences: list[str],
        semantics: list,
        # speaker: Optional[str] = None,
        skip_text: bool = False,
    ):
        # if speaker is None:
        #     speaker = "assistant"

        messages = [
            Message(
                role="system",
                parts=[TextPart(text="Speak out the provided text.")],
                # add_im_end=False,
                # cal_loss=True,
            )
        ]

        cated_sentences = " ".join(sentences)
        if skip_text:
            cated_sentences = "<|skip_text|>"

        messages.append(
            Message(
                role="user",
                parts=[TextPart(text=cated_sentences)],
                # cal_loss=True,
            )
        )

        vq_codes = [x.values for x in semantics[0]]
        vq_codes_tensor = torch.tensor(vq_codes).to(torch.int32)
        vqpart = VQPart(codes=vq_codes_tensor)
        messages.append(
            Message(
                role="assistant",
                parts=[TextPart(text="<|voice|>"), vqpart],
                cal_loss=True,
            )
        )

        num_codebooks = (
            len(semantics[0]) if self.num_codebooks is None else self.num_codebooks
        )

        conversation = Conversation(messages=messages)
        # conversation.visualize(tokenizer=self.tokenizer)
        encoded = conversation.encode(
            tokenizer=self.tokenizer,
        )

        tokens_raw = encoded.tokens
        tokens = torch.zeros((num_codebooks + 1, len(tokens_raw)), dtype=torch.int)
        tokens[0] = tokens_raw

        vq_parts = encoded.vq_parts
        vq_parts = [part.to(tokens.device) for part in vq_parts]
        vq_parts = torch.cat(vq_parts, dim=1)
        tokens[1:, encoded.vq_mask_tokens] = vq_parts

        labels_raw = encoded.labels
        labels = torch.full((num_codebooks + 1, len(labels_raw)), -100, dtype=torch.int)
        labels[0, :] = labels_raw
        labels[1:, encoded.vq_mask_labels] = vq_parts
        labels[1:, -1:] = CODEBOOK_PAD_TOKEN_ID

        tokens = tokens.long()
        labels = labels.long()

        # Verify the padding is correct, and the last token is eos
        assert (tokens[1:, ~(encoded.vq_mask_tokens)] == CODEBOOK_PAD_TOKEN_ID).all()
        assert (labels[1:, -1:] == CODEBOOK_PAD_TOKEN_ID).all()

        return tokens, labels

    def augment(self):
        response = self.sample_data()
        if len(response.samples) == 0:
            # Invalid group
            return None

        samples = list(response.samples)
        all_tokens, all_labels = [], []

        while len(samples) > 0:
            sentence = samples.pop(0)
            text = clean_text(random.choice(sentence.texts))

            tokens, labels = self.pack_sentences(
                sentences=[text],
                semantics=[sentence.semantics],
                # speaker=response.name if use_speaker else None,
                skip_text=random.random() < self.skip_text_prob,
            )

            all_tokens.append(tokens)
            all_labels.append(labels)

        tokens = torch.cat(all_tokens, dim=1)
        labels = torch.cat(all_labels, dim=1)

        # Verify that the length is correct
        assert tokens.size(1) == labels.size(1), f"{tokens.size(1)} != {labels.size(1)}"

        data = {"tokens": tokens, "labels": labels}

        return data


class AutoTextSemanticInstructionDataset(Dataset):
    """
    Auto Augment Dataset by Speaker

    1. Random concatenate multiple sentences from the same speaker to form a longer sentence
    2. Automatically normalize the text

    For interactive mode, we use the following format (multiple sequences):
    <s> [INST] [SPK: speaker] text [/INST] ... [INST] text [/INST] </s>

    For non-interactive mode, we use the following format (one long sequence):
    <s> [INST] text [/INST] ... </s>
    """

    def __init__(
        self,
        proto_files: list[str],
        seed: int = 42,
        interactive_prob: float = 0.5,
        max_length: int = 1024,
        tokenizer: FishTokenizer = None,
        use_speaker: bool | float = True,
        causal: bool = True,
        num_codebooks: Optional[int] = None,
        skip_text_prob: float = 0.0,
    ):
        """
        Args:
            proto_files: proto buf files if using local data
            seed: random seed
            interactive_prob: probability to use interactive mode
            max_length: max length of the text
            tokenizer: tokenizer
            use_speaker: include speaker information in the prompt
            causal: use causal sampling when using local data, disable will lead to random sampling
            num_codebooks: number of codebooks, if None, it will be automatically detected
            skip_text_prob: probability to skip the text (audio only), this only applies to interactive mode
        """
        super().__init__()

        assert 0 <= interactive_prob <= 1, "interactive_prob must be in [0, 1]"

        self.seed = seed
        self.max_length = max_length
        self.tokenizer = tokenizer
        self.interactive_prob = interactive_prob
        self.use_speaker = use_speaker
        self.proto_files = proto_files
        self.causal = causal
        self.num_codebooks = num_codebooks
        self.skip_text_prob = skip_text_prob

        self.data = []
        self._init_data()

    def _init_data(self):
        expanded_proto_files = []
        for filename in self.proto_files:
            for i in braceexpand(filename):
                i = Path(i)
                if i.is_file():
                    expanded_proto_files.append(i)
                elif i.is_dir():
                    expanded_proto_files.extend(i.rglob("*.proto"))
                    expanded_proto_files.extend(i.rglob("*.protos"))
                else:
                    raise ValueError(f"{i} is not a file or directory")

        expanded_proto_files = sorted(expanded_proto_files)
        Random(self.seed).shuffle(expanded_proto_files)

        groups = []
        shard_proto_files = split_by_rank_worker(expanded_proto_files)
        log.info(
            f"Reading {len(shard_proto_files)} / {len(expanded_proto_files)} files"
        )

        count = 0
        for filename in shard_proto_files:
            with open(filename, "rb") as f:
                for text_data in read_pb_stream(f):
                    groups.append(text_data)
                    count += 1

        log.info(f"Read total {count} groups of data")

        for group in groups:
            if len(group.sentences) == 0:
                continue

            samples = list(group.sentences)
            for sentence in samples:
                text = clean_text(random.choice(sentence.texts))

                tokens, labels = self.pack_sentences(
                    sentences=[text],
                    semantics=[sentence.semantics],
                    skip_text=random.random() < self.skip_text_prob,
                )

                self.data.append({"tokens": tokens, "labels": labels})

        random.Random(self.seed).shuffle(self.data)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        return self.data[idx]

    def pack_sentences(
        self,
        sentences: list[str],
        semantics: list,
        skip_text: bool = False,
    ):
        messages = [
            Message(
                role="system",
                parts=[TextPart(text="Speak out the provided text.")],
            )
        ]

        cated_sentences = " ".join(sentences)
        if skip_text:
            cated_sentences = "<|skip_text|>"

        messages.append(
            Message(
                role="user",
                parts=[TextPart(text=cated_sentences)],
            )
        )

        vq_codes = [x.values for x in semantics[0]]
        vq_codes_tensor = torch.tensor(vq_codes).to(torch.int32)
        vqpart = VQPart(codes=vq_codes_tensor)
        messages.append(
            Message(
                role="assistant",
                parts=[TextPart(text="<|voice|>"), vqpart],
                cal_loss=True,
            )
        )

        num_codebooks = (
            len(semantics[0]) if self.num_codebooks is None else self.num_codebooks
        )

        conversation = Conversation(messages=messages)
        encoded = conversation.encode(
            tokenizer=self.tokenizer,
        )

        tokens_raw = encoded.tokens
        tokens = torch.zeros((num_codebooks + 1, len(tokens_raw)), dtype=torch.int)
        tokens[0] = tokens_raw

        vq_parts = encoded.vq_parts
        vq_parts = [part.to(tokens.device) for part in vq_parts]
        vq_parts = torch.cat(vq_parts, dim=1)
        tokens[1:, encoded.vq_mask_tokens] = vq_parts

        labels_raw = encoded.labels
        labels = torch.full((num_codebooks + 1, len(labels_raw)), -100, dtype=torch.int)
        labels[0, :] = labels_raw
        labels[1:, encoded.vq_mask_labels] = vq_parts
        labels[1:, -1:] = CODEBOOK_PAD_TOKEN_ID

        tokens = tokens.long()
        labels = labels.long()

        assert (tokens[1:, ~(encoded.vq_mask_tokens)] == CODEBOOK_PAD_TOKEN_ID).all()
        assert (labels[1:, -1:] == CODEBOOK_PAD_TOKEN_ID).all()

        return tokens, labels


class InterleaveDataset(IterableDataset):
    def __init__(
        self,
        datasets: list[IterableDataset],
        probabilities: list[float],
        seed: int = 42,
    ):
        super().__init__()

        self.datasets = datasets
        self.probabilities = probabilities
        self.seed = seed

    def __iter__(self):
        rng = np.random.default_rng(self.seed)
        dataset_iterators = [iter(dataset) for dataset in self.datasets]

        while True:
            # Random choice one
            dataset_idx = rng.choice(len(self.datasets), p=self.probabilities)
            dataset_iterator = dataset_iterators[dataset_idx]

            try:
                yield next(dataset_iterator)
            except StopIteration:
                # Exhausted, create a new iterator
                dataset_iterators[dataset_idx] = iter(self.datasets[dataset_idx])
                yield next(dataset_iterators[dataset_idx])


@dataclass
class TextDataCollator:
    tokenizer: FishTokenizer
    max_length: int = 1024

    def __call__(self, examples):
        if "negative_tokens" in examples:
            positive_examples = []
            negative_examples = []

            for i in examples:
                positive_examples.append(
                    {
                        "tokens": i["tokens"],
                        "labels": i["labels"],
                    }
                )
                negative_examples.append(
                    {
                        "tokens": i["negative_tokens"],
                        "labels": i["negative_labels"],
                    }
                )

            examples = positive_examples + negative_examples

        return self.batchify(examples)

    def batchify(self, examples, tokens_key="tokens", labels_key="labels"):
        tokens, attention_masks, labels = [], [], []

        # Calculate the max length
        max_tokens_length = 0
        for example in examples:
            max_tokens_length = max(max_tokens_length, example[tokens_key].size(1))
        max_tokens_length = min(max_tokens_length, self.max_length)

        for example in examples:
            _tokens = example[tokens_key][:, :max_tokens_length]
            _labels = example[labels_key][:, :max_tokens_length]
            _attention_mask = torch.ones((max_tokens_length,), dtype=torch.bool)
            tokens_length = _tokens.size(1)
            _attention_mask[:tokens_length] = False

            assert tokens_length == _labels.size(
                1
            ), f"{tokens_length} != {_labels.size(1)}"

            if tokens_length < max_tokens_length:
                _tokens = F.pad(
                    _tokens,
                    (0, max_tokens_length - tokens_length),
                    value=self.tokenizer.get_token_id("<|end_of_text|>"),
                )
                _tokens[1:, tokens_length:] = CODEBOOK_PAD_TOKEN_ID
                _labels = F.pad(
                    _labels, (0, max_tokens_length - _labels.size(1)), value=-100
                )

            tokens.append(_tokens)
            attention_masks.append(_attention_mask)
            labels.append(_labels)

        tokens = torch.stack(tokens, dim=0)
        attention_masks = torch.stack(attention_masks, dim=0)
        labels = torch.stack(labels, dim=0)

        return {
            "inputs": tokens,
            "attention_masks": attention_masks,
            "labels": labels,
        }


class SemanticDataModule(LightningDataModule):
    def __init__(
        self,
        train_dataset: Union[
            AutoTextSemanticInstructionDataset,
            AutoTextSemanticInstructionIterableDataset,
            InterleaveDataset,
        ],
        val_dataset: Union[
            AutoTextSemanticInstructionDataset,
            AutoTextSemanticInstructionIterableDataset,
            InterleaveDataset,
        ],
        batch_size: int = 32,
        tokenizer: FishTokenizer = None,
        max_length: int = 1024,
        num_workers: int = 4,
    ):
        super().__init__()

        self.train_dataset = train_dataset
        self.val_dataset = val_dataset
        self.batch_size = batch_size
        self.tokenizer = tokenizer
        self.max_length = max_length
        self.num_workers = num_workers

    def train_dataloader(self):
        return DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,
            collate_fn=TextDataCollator(self.tokenizer, self.max_length),
            num_workers=self.num_workers,
            persistent_workers=True,
        )

    def val_dataloader(self):
        return DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,
            collate_fn=TextDataCollator(self.tokenizer, self.max_length),
            num_workers=self.num_workers,
            persistent_workers=True,
        )


if __name__ == "__main__":
    from tqdm import tqdm

    ds = AutoTextSemanticInstructionDataset(
        ["data/protos"],
        tokenizer=FishTokenizer("checkpoints/fish-speech-1.5/tokenizer.tiktoken"),
        use_speaker=False,
        interactive_prob=1.0,
        skip_text_prob=0.5,
    )

    for i in range(100):
        # Please uncomment line 235 to visualize the tokenized message
        print(ds[i])



# just for speaker similarity evaluation, third-party code

# From https://github.com/microsoft/UniSpeech/blob/main/downstreams/speaker_verification/models/
# part of the code is borrowed from https://github.com/lawlict/ECAPA-TDNN

import os
import torch
import torch.nn as nn
import torch.nn.functional as F


""" Res2Conv1d + BatchNorm1d + ReLU
"""


class Res2Conv1dReluBn(nn.Module):
    """
    in_channels == out_channels == channels
    """

    def __init__(self, channels, kernel_size=1, stride=1, padding=0, dilation=1, bias=True, scale=4):
        super().__init__()
        assert channels % scale == 0, "{} % {} != 0".format(channels, scale)
        self.scale = scale
        self.width = channels // scale
        self.nums = scale if scale == 1 else scale - 1

        self.convs = []
        self.bns = []
        for i in range(self.nums):
            self.convs.append(nn.Conv1d(self.width, self.width, kernel_size, stride, padding, dilation, bias=bias))
            self.bns.append(nn.BatchNorm1d(self.width))
        self.convs = nn.ModuleList(self.convs)
        self.bns = nn.ModuleList(self.bns)

    def forward(self, x):
        out = []
        spx = torch.split(x, self.width, 1)
        for i in range(self.nums):
            if i == 0:
                sp = spx[i]
            else:
                sp = sp + spx[i]
            # Order: conv -> relu -> bn
            sp = self.convs[i](sp)
            sp = self.bns[i](F.relu(sp))
            out.append(sp)
        if self.scale != 1:
            out.append(spx[self.nums])
        out = torch.cat(out, dim=1)

        return out


""" Conv1d + BatchNorm1d + ReLU
"""


class Conv1dReluBn(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size=1, stride=1, padding=0, dilation=1, bias=True):
        super().__init__()
        self.conv = nn.Conv1d(in_channels, out_channels, kernel_size, stride, padding, dilation, bias=bias)
        self.bn = nn.BatchNorm1d(out_channels)

    def forward(self, x):
        return self.bn(F.relu(self.conv(x)))


""" The SE connection of 1D case.
"""


class SE_Connect(nn.Module):
    def __init__(self, channels, se_bottleneck_dim=128):
        super().__init__()
        self.linear1 = nn.Linear(channels, se_bottleneck_dim)
        self.linear2 = nn.Linear(se_bottleneck_dim, channels)

    def forward(self, x):
        out = x.mean(dim=2)
        out = F.relu(self.linear1(out))
        out = torch.sigmoid(self.linear2(out))
        out = x * out.unsqueeze(2)

        return out


""" SE-Res2Block of the ECAPA-TDNN architecture.
"""

# def SE_Res2Block(channels, kernel_size, stride, padding, dilation, scale):
#     return nn.Sequential(
#         Conv1dReluBn(channels, 512, kernel_size=1, stride=1, padding=0),
#         Res2Conv1dReluBn(512, kernel_size, stride, padding, dilation, scale=scale),
#         Conv1dReluBn(512, channels, kernel_size=1, stride=1, padding=0),
#         SE_Connect(channels)
#     )


class SE_Res2Block(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, stride, padding, dilation, scale, se_bottleneck_dim):
        super().__init__()
        self.Conv1dReluBn1 = Conv1dReluBn(in_channels, out_channels, kernel_size=1, stride=1, padding=0)
        self.Res2Conv1dReluBn = Res2Conv1dReluBn(out_channels, kernel_size, stride, padding, dilation, scale=scale)
        self.Conv1dReluBn2 = Conv1dReluBn(out_channels, out_channels, kernel_size=1, stride=1, padding=0)
        self.SE_Connect = SE_Connect(out_channels, se_bottleneck_dim)

        self.shortcut = None
        if in_channels != out_channels:
            self.shortcut = nn.Conv1d(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=1,
            )

    def forward(self, x):
        residual = x
        if self.shortcut:
            residual = self.shortcut(x)

        x = self.Conv1dReluBn1(x)
        x = self.Res2Conv1dReluBn(x)
        x = self.Conv1dReluBn2(x)
        x = self.SE_Connect(x)

        return x + residual


""" Attentive weighted mean and standard deviation pooling.
"""


class AttentiveStatsPool(nn.Module):
    def __init__(self, in_dim, attention_channels=128, global_context_att=False):
        super().__init__()
        self.global_context_att = global_context_att

        # Use Conv1d with stride == 1 rather than Linear, then we don't need to transpose inputs.
        if global_context_att:
            self.linear1 = nn.Conv1d(in_dim * 3, attention_channels, kernel_size=1)  # equals W and b in the paper
        else:
            self.linear1 = nn.Conv1d(in_dim, attention_channels, kernel_size=1)  # equals W and b in the paper
        self.linear2 = nn.Conv1d(attention_channels, in_dim, kernel_size=1)  # equals V and k in the paper

    def forward(self, x):
        if self.global_context_att:
            context_mean = torch.mean(x, dim=-1, keepdim=True).expand_as(x)
            context_std = torch.sqrt(torch.var(x, dim=-1, keepdim=True) + 1e-10).expand_as(x)
            x_in = torch.cat((x, context_mean, context_std), dim=1)
        else:
            x_in = x

        # DON'T use ReLU here! In experiments, I find ReLU hard to converge.
        alpha = torch.tanh(self.linear1(x_in))
        # alpha = F.relu(self.linear1(x_in))
        alpha = torch.softmax(self.linear2(alpha), dim=2)
        mean = torch.sum(alpha * x, dim=2)
        residuals = torch.sum(alpha * (x**2), dim=2) - mean**2
        std = torch.sqrt(residuals.clamp(min=1e-9))
        return torch.cat([mean, std], dim=1)


class ECAPA_TDNN(nn.Module):
    def __init__(
        self,
        feat_dim=80,
        channels=512,
        emb_dim=192,
        global_context_att=False,
        feat_type="wavlm_large",
        sr=16000,
        feature_selection="hidden_states",
        update_extract=False,
        config_path=None,
    ):
        super().__init__()

        self.feat_type = feat_type
        self.feature_selection = feature_selection
        self.update_extract = update_extract
        self.sr = sr

        torch.hub._validate_not_a_forked_repo = lambda a, b, c: True
        try:
            local_s3prl_path = os.path.expanduser("~/.cache/torch/hub/s3prl_s3prl_main")
            self.feature_extract = torch.hub.load(local_s3prl_path, feat_type, source="local", config_path=config_path)
        except:  # noqa: E722
            self.feature_extract = torch.hub.load("s3prl/s3prl", feat_type)

        if len(self.feature_extract.model.encoder.layers) == 24 and hasattr(
            self.feature_extract.model.encoder.layers[23].self_attn, "fp32_attention"
        ):
            self.feature_extract.model.encoder.layers[23].self_attn.fp32_attention = False
        if len(self.feature_extract.model.encoder.layers) == 24 and hasattr(
            self.feature_extract.model.encoder.layers[11].self_attn, "fp32_attention"
        ):
            self.feature_extract.model.encoder.layers[11].self_attn.fp32_attention = False

        self.feat_num = self.get_feat_num()
        self.feature_weight = nn.Parameter(torch.zeros(self.feat_num))

        if feat_type != "fbank" and feat_type != "mfcc":
            freeze_list = ["final_proj", "label_embs_concat", "mask_emb", "project_q", "quantizer"]
            for name, param in self.feature_extract.named_parameters():
                for freeze_val in freeze_list:
                    if freeze_val in name:
                        param.requires_grad = False
                        break

        if not self.update_extract:
            for param in self.feature_extract.parameters():
                param.requires_grad = False

        self.instance_norm = nn.InstanceNorm1d(feat_dim)
        # self.channels = [channels] * 4 + [channels * 3]
        self.channels = [channels] * 4 + [1536]

        self.layer1 = Conv1dReluBn(feat_dim, self.channels[0], kernel_size=5, padding=2)
        self.layer2 = SE_Res2Block(
            self.channels[0],
            self.channels[1],
            kernel_size=3,
            stride=1,
            padding=2,
            dilation=2,
            scale=8,
            se_bottleneck_dim=128,
        )
        self.layer3 = SE_Res2Block(
            self.channels[1],
            self.channels[2],
            kernel_size=3,
            stride=1,
            padding=3,
            dilation=3,
            scale=8,
            se_bottleneck_dim=128,
        )
        self.layer4 = SE_Res2Block(
            self.channels[2],
            self.channels[3],
            kernel_size=3,
            stride=1,
            padding=4,
            dilation=4,
            scale=8,
            se_bottleneck_dim=128,
        )

        # self.conv = nn.Conv1d(self.channels[-1], self.channels[-1], kernel_size=1)
        cat_channels = channels * 3
        self.conv = nn.Conv1d(cat_channels, self.channels[-1], kernel_size=1)
        self.pooling = AttentiveStatsPool(
            self.channels[-1], attention_channels=128, global_context_att=global_context_att
        )
        self.bn = nn.BatchNorm1d(self.channels[-1] * 2)
        self.linear = nn.Linear(self.channels[-1] * 2, emb_dim)

    def get_feat_num(self):
        self.feature_extract.eval()
        wav = [torch.randn(self.sr).to(next(self.feature_extract.parameters()).device)]
        with torch.no_grad():
            features = self.feature_extract(wav)
        select_feature = features[self.feature_selection]
        if isinstance(select_feature, (list, tuple)):
            return len(select_feature)
        else:
            return 1

    def get_feat(self, x):
        if self.update_extract:
            x = self.feature_extract([sample for sample in x])
        else:
            with torch.no_grad():
                if self.feat_type == "fbank" or self.feat_type == "mfcc":
                    x = self.feature_extract(x) + 1e-6  # B x feat_dim x time_len
                else:
                    x = self.feature_extract([sample for sample in x])

        if self.feat_type == "fbank":
            x = x.log()

        if self.feat_type != "fbank" and self.feat_type != "mfcc":
            x = x[self.feature_selection]
            if isinstance(x, (list, tuple)):
                x = torch.stack(x, dim=0)
            else:
                x = x.unsqueeze(0)
            norm_weights = F.softmax(self.feature_weight, dim=-1).unsqueeze(-1).unsqueeze(-1).unsqueeze(-1)
            x = (norm_weights * x).sum(dim=0)
            x = torch.transpose(x, 1, 2) + 1e-6

        x = self.instance_norm(x)
        return x

    def forward(self, x):
        x = self.get_feat(x)

        out1 = self.layer1(x)
        out2 = self.layer2(out1)
        out3 = self.layer3(out2)
        out4 = self.layer4(out3)

        out = torch.cat([out2, out3, out4], dim=1)
        out = F.relu(self.conv(out))
        out = self.bn(self.pooling(out))
        out = self.linear(out)

        return out


def ECAPA_TDNN_SMALL(
    feat_dim,
    emb_dim=256,
    feat_type="wavlm_large",
    sr=16000,
    feature_selection="hidden_states",
    update_extract=False,
    config_path=None,
):
    return ECAPA_TDNN(
        feat_dim=feat_dim,
        channels=512,
        emb_dim=emb_dim,
        feat_type=feat_type,
        sr=sr,
        feature_selection=feature_selection,
        update_extract=update_extract,
        config_path=config_path,
    )



# Copyright 2025, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#  * Neither the name of NVIDIA CORPORATION nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
# OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import json
import torch
from torch.utils.dlpack import to_dlpack

import triton_python_backend_utils as pb_utils

import os
import numpy as np

from sparktts.models.audio_tokenizer import BiCodecTokenizer

class TritonPythonModel:
    """Triton Python model for audio tokenization.
    
    This model takes reference audio input and extracts semantic and global tokens
    using BiCodec tokenizer.
    """

    def initialize(self, args):
        """Initialize the model.
        
        Args:
            args: Dictionary containing model configuration
        """
        # Parse model parameters
        parameters = json.loads(args['model_config'])['parameters']
        model_params = {k: v["string_value"] for k, v in parameters.items()}
        
        # Initialize tokenizer
        self.device = torch.device("cuda")
        self.audio_tokenizer = BiCodecTokenizer(model_params["model_dir"], 
                                              device=self.device)

    def get_ref_clip(self, wav: np.ndarray) -> np.ndarray:
        """Extract reference audio clip for speaker embedding.
        
        Args:
            wav: Input waveform array
            
        Returns:
            Reference clip of fixed duration
        """
        SAMPLE_RATE = 16000
        REF_SEGMENT_DURATION = 6  # seconds
        LATENT_HOP_LENGTH = 320

        ref_segment_length = (
            int(SAMPLE_RATE * REF_SEGMENT_DURATION)
            // LATENT_HOP_LENGTH
            * LATENT_HOP_LENGTH
        )
        wav_length = len(wav)

        if ref_segment_length > wav_length:
            # Repeat and truncate if input is too short
            repeat_times = ref_segment_length // wav_length + 1
            wav = np.tile(wav, repeat_times)

        return wav[:ref_segment_length]

    def execute(self, requests):
        """Execute inference on the batched requests.
        
        Args:
            requests: List of inference requests
            
        Returns:
            List of inference responses containing tokenized outputs
        """
        reference_wav_list = []
        reference_wav_ref_clip_list = []

        # Process each request in batch
        for request in requests:
            # Extract input tensors
            wav_array = pb_utils.get_input_tensor_by_name(
                request, "reference_wav").as_numpy()
            wav_len = pb_utils.get_input_tensor_by_name(
                request, "reference_wav_len").as_numpy().item()

            # Prepare inputs
            wav = wav_array[:, :wav_len].squeeze(0)
            reference_wav_list.append(wav)
            
            wav_ref_clip = self.get_ref_clip(wav)
            reference_wav_ref_clip_list.append(torch.from_numpy(wav_ref_clip))

        # Batch process through tokenizer
        ref_wav_clip_tensor = torch.stack(reference_wav_ref_clip_list, dim=0)
        wav2vec2_features = self.audio_tokenizer.extract_wav2vec2_features(
            reference_wav_list)
        
        audio_tokenizer_input = {
            "ref_wav": ref_wav_clip_tensor.to(self.device),
            "feat": wav2vec2_features.to(self.device),
        }
        semantic_tokens, global_tokens = self.audio_tokenizer.model.tokenize(
            audio_tokenizer_input)

        # Prepare responses
        responses = []
        for i in range(len(requests)):
            global_tokens_tensor = pb_utils.Tensor.from_dlpack(
                "global_tokens", to_dlpack(global_tokens[i]))
            semantic_tokens_tensor = pb_utils.Tensor.from_dlpack(
                "semantic_tokens", to_dlpack(semantic_tokens[i]))
            
            inference_response = pb_utils.InferenceResponse(
                output_tensors=[global_tokens_tensor, semantic_tokens_tensor])
            responses.append(inference_response)
                             
        return responses



# Copyright (c) 2024 Alibaba Inc (authors: Xiang Lyu)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import argparse
import logging
logging.getLogger('matplotlib').setLevel(logging.WARNING)
import os
import torch
from torch.utils.data import DataLoader
import torchaudio
from hyperpyyaml import load_hyperpyyaml
from tqdm import tqdm
from cosyvoice.cli.model import CosyVoiceModel
from cosyvoice.dataset.dataset import Dataset


def get_args():
    parser = argparse.ArgumentParser(description='inference with your model')
    parser.add_argument('--config', required=True, help='config file')
    parser.add_argument('--prompt_data', required=True, help='prompt data file')
    parser.add_argument('--prompt_utt2data', required=True, help='prompt data file')
    parser.add_argument('--tts_text', required=True, help='tts input file')
    parser.add_argument('--llm_model', required=True, help='llm model file')
    parser.add_argument('--flow_model', required=True, help='flow model file')
    parser.add_argument('--hifigan_model', required=True, help='hifigan model file')
    parser.add_argument('--gpu',
                        type=int,
                        default=-1,
                        help='gpu id for this rank, -1 for cpu')
    parser.add_argument('--mode',
                        default='sft',
                        choices=['sft', 'zero_shot'],
                        help='inference mode')
    parser.add_argument('--result_dir', required=True, help='asr result file')
    args = parser.parse_args()
    print(args)
    return args


def main():
    args = get_args()
    logging.basicConfig(level=logging.DEBUG,
                        format='%(asctime)s %(levelname)s %(message)s')
    os.environ['CUDA_VISIBLE_DEVICES'] = str(args.gpu)

    # Init cosyvoice models from configs
    use_cuda = args.gpu >= 0 and torch.cuda.is_available()
    device = torch.device('cuda' if use_cuda else 'cpu')
    with open(args.config, 'r') as f:
        configs = load_hyperpyyaml(f)

    model = CosyVoiceModel(configs['llm'], configs['flow'], configs['hift'])
    model.load(args.llm_model, args.flow_model, args.hifigan_model)

    test_dataset = Dataset(args.prompt_data, data_pipeline=configs['data_pipeline'], mode='inference', shuffle=False, partition=False,
                           tts_file=args.tts_text, prompt_utt2data=args.prompt_utt2data)
    test_data_loader = DataLoader(test_dataset, batch_size=None, num_workers=0)

    del configs
    os.makedirs(args.result_dir, exist_ok=True)
    fn = os.path.join(args.result_dir, 'wav.scp')
    f = open(fn, 'w')
    with torch.no_grad():
        for _, batch in tqdm(enumerate(test_data_loader)):
            utts = batch["utts"]
            assert len(utts) == 1, "inference mode only support batchsize 1"
            text_token = batch["text_token"].to(device)
            text_token_len = batch["text_token_len"].to(device)
            tts_index = batch["tts_index"]
            tts_text_token = batch["tts_text_token"].to(device)
            tts_text_token_len = batch["tts_text_token_len"].to(device)
            speech_token = batch["speech_token"].to(device)
            speech_token_len = batch["speech_token_len"].to(device)
            speech_feat = batch["speech_feat"].to(device)
            speech_feat_len = batch["speech_feat_len"].to(device)
            utt_embedding = batch["utt_embedding"].to(device)
            spk_embedding = batch["spk_embedding"].to(device)
            if args.mode == 'sft':
                model_input = {'text': tts_text_token, 'text_len': tts_text_token_len,
                               'llm_embedding': spk_embedding, 'flow_embedding': spk_embedding}
            else:
                model_input = {'text': tts_text_token, 'text_len': tts_text_token_len,
                               'prompt_text': text_token, 'prompt_text_len': text_token_len,
                               'llm_prompt_speech_token': speech_token, 'llm_prompt_speech_token_len': speech_token_len,
                               'flow_prompt_speech_token': speech_token, 'flow_prompt_speech_token_len': speech_token_len,
                               'prompt_speech_feat': speech_feat, 'prompt_speech_feat_len': speech_feat_len,
                               'llm_embedding': utt_embedding, 'flow_embedding': utt_embedding}
            tts_speeches = []
            for model_output in model.tts(**model_input):
                tts_speeches.append(model_output['tts_speech'])
            tts_speeches = torch.concat(tts_speeches, dim=1)
            tts_key = '{}_{}'.format(utts[0], tts_index[0])
            tts_fn = os.path.join(args.result_dir, '{}.wav'.format(tts_key))
            torchaudio.save(tts_fn, tts_speeches, sample_rate=22050)
            f.write('{} {}\n'.format(tts_key, tts_fn))
            f.flush()
    f.close()
    logging.info('Result wav.scp saved in {}'.format(fn))


if __name__ == '__main__':
    main()



import io
import logging
import os
import sys
from threading import Thread
from dotenv import load_dotenv

from PIL import Image
import numpy as np

from src.common.chat_history import ChatHistory
from src.common.session import Session
from src.common.utils.helper import get_device, print_model_params
from src.core.llm.transformers.base import TransformersBaseLLM
from src.types.llm.transformers import TransformersLMArgs
from src.common.random import set_all_random_seed

load_dotenv(override=True)

try:
    cur_dir = os.path.dirname(__file__)
    if bool(os.getenv("ACHATBOT_PKG", "")):
        sys.path.insert(1, os.path.join(cur_dir, "../../../DeepSeekVL2"))
    else:
        sys.path.insert(1, os.path.join(cur_dir, "../../../../deps/DeepSeekVL2"))
    import torch
    from transformers import AutoModelForCausalLM, AutoConfig, TextIteratorStreamer
    from deps.DeepSeekVL2.deepseek_vl2.models import DeepseekVLV2Processor, DeepseekVLV2ForCausalLM

except ModuleNotFoundError as e:
    logging.error(
        "In order to use DeepSeek VL2, you need to `pip install achatbot[llm_transformers_manual_vision_deepseekvl2]`."
    )
    raise Exception(f"Missing module: {e}")


def split_model(model_name):
    device_map = {}

    # splits layers into different GPUs (need use L4 for bfloat16 with flash attention)
    model_splits = {
        # DeepSeek-VL2-tiny is built on DeepSeekMoE-3B (total activated parameters are 1.0B) with 12 layers
        "deepseek-ai/deepseek-vl2-tiny": [12],  # 1 GPU for 3B
        # DeepSeek-VL2-small is built on DeepSeekMoE-16B (total activated parameters are 2.8B) with 27 layers
        "deepseek-ai/deepseek-vl2-small": [13, 14],  # 2 GPU for 16b
        # DeepSeek-VL2 is built on DeepSeekMoE-27B (total activated parameters are 4.5B) with 30 layers
        "deepseek-ai/deepseek-vl2": [6, 8, 8, 8],  # 4 GPU for 27b
    }
    num_layers_per_gpu = model_splits[model_name]
    num_layers = sum(num_layers_per_gpu)
    layer_cnt = 0
    for i, num_layer in enumerate(num_layers_per_gpu):
        for j in range(num_layer):
            device_map[f"language.model.layers.{layer_cnt}"] = i
            layer_cnt += 1

    # exlude layer and last layer on cuda 0
    device_map["vision"] = 0
    device_map["projector"] = 0
    device_map["image_newline"] = 0
    device_map["view_seperator"] = 0
    device_map["language.model.embed_tokens"] = 0
    device_map["language.model.norm"] = 0
    device_map["language.lm_head"] = 0
    device_map[f"language.model.layers.{num_layers - 1}"] = 0
    return device_map


class TransformersManualVisionDeepSeekVL2(TransformersBaseLLM):
    r"""
    Multimodal Understanding
    https://github.com/deepseek-ai/DeepSeek-VL2

    vl_chat_processor.tokenizer.encode + AR LM model(MHA/MLA+MoE) + vl_chat_processor.tokenizer.decode
    """

    TAG = "llm_transformers_manual_vision_deepseek"

    def __init__(self, **args) -> None:
        self.args = TransformersLMArgs(**args)

        # https://huggingface.co/deepseek-ai/deepseek-vl2/blob/main/config.json (MLA + MOE)
        # https://huggingface.co/deepseek-ai/deepseek-vl2-small/blob/main/config.json (MLA + MOE)
        # https://huggingface.co/deepseek-ai/deepseek-vl2-tiny/blob/main/config.json (MHA + MOE)
        config = AutoConfig.from_pretrained(self.args.lm_model_name_or_path)
        language_config = config.language_config
        language_config._attn_implementation = "eager"
        if self.args.lm_device_map is not None:
            if isinstance(self.args.lm_device_map, dict):
                customer_deivce_map = self.args.lm_device_map
                default_device_map = split_model(
                    "/".join(self.args.lm_model_name_or_path.split("/")[-2:])
                )
                self.args.lm_device_map = {**default_device_map, **customer_deivce_map}
            logging.info(f"TransformersLMArgs: {self.args}")
            self._model: DeepseekVLV2ForCausalLM = AutoModelForCausalLM.from_pretrained(
                self.args.lm_model_name_or_path,
                language_config=language_config,
                trust_remote_code=True,
                torch_dtype=torch.bfloat16 if torch.cuda.is_available() else torch.float16,
                device_map=self.args.lm_device_map,
            ).eval()
        else:
            self.args.lm_device = self.args.lm_device or get_device()
            logging.info(f"TransformersLMArgs: {self.args}")
            self._model: DeepseekVLV2ForCausalLM = AutoModelForCausalLM.from_pretrained(
                self.args.lm_model_name_or_path,
                language_config=language_config,
                trust_remote_code=True,
            )
            self._model = self._model.to(
                self.args.lm_device,
                dtype=torch.bfloat16 if torch.cuda.is_available() else torch.float16,
            ).eval()
        logging.info(f"TransformersLMArgs: {self.args} model.device: {self._model.device}")
        print_model_params(self._model, self.TAG)

        self.vl_chat_processor: DeepseekVLV2Processor = DeepseekVLV2Processor.from_pretrained(
            self.args.lm_model_name_or_path
        )
        self._tokenizer = self.vl_chat_processor.tokenizer

        self._streamer = TextIteratorStreamer(
            self._tokenizer, skip_prompt=True, skip_special_tokens=True
        )

        self._chat_history = ChatHistory(self.args.chat_history_size)
        self.warmup()

    def warmup(self):
        dummy_input_text = self.args.warnup_prompt
        conversation = [
            {
                "role": "<|User|>",
                "content": f"<image>\n{dummy_input_text}",
            },
            {"role": "<|Assistant|>", "content": ""},
        ]

        dummy_pil_images = [Image.new("RGB", (100, 100), color="white")]
        prepare_inputs = self.vl_chat_processor(
            conversations=conversation,
            images=dummy_pil_images,
            force_batchify=True,
            system_prompt="",
        ).to(
            self._model.device,
            dtype=torch.bfloat16 if torch.cuda.is_available() else torch.float16,
        )
        # logging.debug(f"prepare_inputs: {prepare_inputs}")

        # input embeddings
        inputs_embeds = self._model.prepare_inputs_embeds(**prepare_inputs)
        # AR lm generate with streamer
        warmup_gen_kwargs = dict(
            inputs_embeds=inputs_embeds,
            streamer=self._streamer,
            attention_mask=prepare_inputs.attention_mask,
            pad_token_id=self._tokenizer.eos_token_id,
            bos_token_id=self._tokenizer.bos_token_id,
            eos_token_id=self._tokenizer.eos_token_id,
            max_new_tokens=self.args.lm_gen_max_new_tokens,
            do_sample=False if self.args.lm_gen_temperature == 0 else True,
            use_cache=True,
            temperature=self.args.lm_gen_temperature,
            top_p=self.args.lm_gen_top_p,
        )

        self._warmup(
            target=self._model.language.generate,
            kwargs=warmup_gen_kwargs,
            streamer=self._streamer,
        )

    @torch.inference_mode()
    def generate(self, session: Session, **kwargs):
        logging.debug(f"kwargs: {kwargs}")
        if "cuda" in str(self._model.device):
            torch.cuda.empty_cache()
        seed = kwargs.get("seed", self.args.lm_gen_seed)
        set_all_random_seed(seed)

        assert isinstance(session.ctx.state["prompt"], list)
        assert len(session.ctx.state["prompt"]) > 1
        question = session.ctx.state["prompt"][-1]
        pil_images = session.ctx.state["prompt"][:-1]

        lm_gen_max_new_tokens = kwargs.get("lm_gen_max_new_tokens", self.args.lm_gen_max_new_tokens)
        lm_gen_temperature = kwargs.get("lm_gen_temperature", self.args.lm_gen_temperature)
        lm_gen_top_p = kwargs.get("lm_gen_top_p", self.args.lm_gen_top_p)

        # conversation
        message = {
            "role": "<|User|>",
            "content": f"<image>\n{question}",
            # "images": [image_data],
        }
        self._chat_history.append(message)
        chat_history = self._chat_history.to_list()
        logging.debug(f"chat_history:{chat_history}")
        conversation = chat_history + [{"role": "<|Assistant|>", "content": ""}]

        # inputs
        # pil_images = [Image.open(io.BytesIO(image_data))]
        prepare_inputs = self.vl_chat_processor(
            conversations=conversation, images=pil_images, force_batchify=True
        ).to(
            self._model.device,
            dtype=torch.bfloat16 if torch.cuda.is_available() else torch.float16,
        )
        # logging.debug(f"prepare_inputs: {prepare_inputs}")

        # input embeddings
        inputs_embeds = self._model.prepare_inputs_embeds(**prepare_inputs)

        # AR lm generate with streamer
        generation_kwargs = dict(
            inputs_embeds=inputs_embeds,
            streamer=self._streamer,
            attention_mask=prepare_inputs.attention_mask,
            pad_token_id=self._tokenizer.eos_token_id,
            bos_token_id=self._tokenizer.bos_token_id,
            eos_token_id=self._tokenizer.eos_token_id,
            max_new_tokens=lm_gen_max_new_tokens,
            do_sample=False if lm_gen_temperature == 0 else True,
            use_cache=True,
            temperature=lm_gen_temperature,
            top_p=lm_gen_top_p,
        )
        thread = Thread(target=self._model.language.generate, kwargs=generation_kwargs)
        thread.start()

        generated_text = ""
        for new_text in self._streamer:
            generated_text += new_text
            yield new_text
        self._chat_history.append({"role": "<|Assistant|>", "content": generated_text})

        torch.cuda.empty_cache()



import re

SYMBOLS_MAPPING = {
    "‘": "'",
    "’": "'",
}

REPLACE_SYMBOL_REGEX = re.compile(
    "|".join(re.escape(p) for p in SYMBOLS_MAPPING.keys())
)


EMOJI_REGEX = re.compile(
    "["
    "\U0001F600-\U0001F64F"  # emoticons
    "\U0001F300-\U0001F5FF"  # symbols & pictographs
    "\U0001F680-\U0001F6FF"  # transport & map symbols
    "\U0001F1E0-\U0001F1FF"  # flags (iOS)
    "]+",
    flags=re.UNICODE,
)


def clean_text(text):
    # Clean the text
    text = text.strip()

    # Replace all chinese symbols with their english counterparts
    text = REPLACE_SYMBOL_REGEX.sub(lambda x: SYMBOLS_MAPPING[x.group()], text)

    # Remove emojis
    text = EMOJI_REGEX.sub(r"", text)

    # Remove continuous periods (...) and commas (,,,)
    text = re.sub(r"[,]{2,}", lambda m: m.group()[0], text)

    return text



import time
import asyncio


class Pacer:
    def __init__(self, interval):
        self.last_call_time = time.time()
        self.interval = interval

    def pace(self):
        current_time = time.time()
        elapsed_time = current_time - self.last_call_time
        if elapsed_time < self.interval:
            time.sleep(self.interval - elapsed_time)
            # print("sleep time:", (self.interval - elapsed_time)*1000)
        self.last_call_time = time.time()

    def pace_interval(self, time_interval_s):
        current_time = time.time()
        elapsed_time = current_time - self.last_call_time
        if elapsed_time < time_interval_s:
            time.sleep(time_interval_s - elapsed_time)
            # print("sleep time(ms):", (time_interval_s - elapsed_time)*1000)
        self.last_call_time = time.time()

    async def apace_interval(self, time_interval_s):
        current_time = time.time()
        elapsed_time = current_time - self.last_call_time
        # logging.info(f"elapsed_time:{elapsed_time}, time_interval_s:{time_interval_s}")
        if elapsed_time < time_interval_s:
            await asyncio.sleep(time_interval_s - elapsed_time)
        self.last_call_time = time.time()



import argparse
import random
from pathlib import Path

import numpy as np
import torch
from lightning import LightningModule

from matcha.cli import VOCODER_URLS, load_matcha, load_vocoder

DEFAULT_OPSET = 15

SEED = 1234
random.seed(SEED)
np.random.seed(SEED)
torch.manual_seed(SEED)
torch.cuda.manual_seed(SEED)
torch.backends.cudnn.deterministic = True
torch.backends.cudnn.benchmark = False


class MatchaWithVocoder(LightningModule):
    def __init__(self, matcha, vocoder):
        super().__init__()
        self.matcha = matcha
        self.vocoder = vocoder

    def forward(self, x, x_lengths, scales, spks=None):
        mel, mel_lengths = self.matcha(x, x_lengths, scales, spks)
        wavs = self.vocoder(mel).clamp(-1, 1)
        lengths = mel_lengths * 256
        return wavs.squeeze(1), lengths


def get_exportable_module(matcha, vocoder, n_timesteps):
    """
    Return an appropriate `LighteningModule` and output-node names
    based on whether the vocoder is embedded in  the final graph
    """

    def onnx_forward_func(x, x_lengths, scales, spks=None):
        """
        Custom forward function for accepting
        scaler parameters as tensors
        """
        # Extract scaler parameters from tensors
        temperature = scales[0]
        length_scale = scales[1]
        output = matcha.synthesise(x, x_lengths, n_timesteps, temperature, spks, length_scale)
        return output["mel"], output["mel_lengths"]

    # Monkey-patch Matcha's forward function
    matcha.forward = onnx_forward_func

    if vocoder is None:
        model, output_names = matcha, ["mel", "mel_lengths"]
    else:
        model = MatchaWithVocoder(matcha, vocoder)
        output_names = ["wav", "wav_lengths"]
    return model, output_names


def get_inputs(is_multi_speaker):
    """
    Create dummy inputs for tracing
    """
    dummy_input_length = 50
    x = torch.randint(low=0, high=20, size=(1, dummy_input_length), dtype=torch.long)
    x_lengths = torch.LongTensor([dummy_input_length])

    # Scales
    temperature = 0.667
    length_scale = 1.0
    scales = torch.Tensor([temperature, length_scale])

    model_inputs = [x, x_lengths, scales]
    input_names = [
        "x",
        "x_lengths",
        "scales",
    ]

    if is_multi_speaker:
        spks = torch.LongTensor([1])
        model_inputs.append(spks)
        input_names.append("spks")

    return tuple(model_inputs), input_names


def main():
    parser = argparse.ArgumentParser(description="Export 🍵 Matcha-TTS to ONNX")

    parser.add_argument(
        "checkpoint_path",
        type=str,
        help="Path to the model checkpoint",
    )
    parser.add_argument("output", type=str, help="Path to output `.onnx` file")
    parser.add_argument(
        "--n-timesteps",
        type=int,
        default=5,
        help="Number of steps to use for reverse diffusion in decoder (default 5)",
    )
    parser.add_argument(
        "--vocoder-name",
        type=str,
        choices=list(VOCODER_URLS.keys()),
        default=None,
        help="Name of the vocoder to embed in the ONNX graph",
    )
    parser.add_argument(
        "--vocoder-checkpoint-path",
        type=str,
        default=None,
        help="Vocoder checkpoint to embed  in the ONNX graph for an `e2e` like experience",
    )
    parser.add_argument(
        "--opset", type=int, default=DEFAULT_OPSET, help="ONNX opset version to use (default 15"
    )

    args = parser.parse_args()

    print(f"[🍵] Loading Matcha checkpoint from {args.checkpoint_path}")
    print(f"Setting n_timesteps to {args.n_timesteps}")

    checkpoint_path = Path(args.checkpoint_path)
    matcha = load_matcha(checkpoint_path.stem, checkpoint_path, "cpu")

    if args.vocoder_name or args.vocoder_checkpoint_path:
        assert (
            args.vocoder_name and args.vocoder_checkpoint_path
        ), "Both vocoder_name and vocoder-checkpoint are required when embedding the vocoder in the ONNX graph."
        vocoder, _ = load_vocoder(args.vocoder_name, args.vocoder_checkpoint_path, "cpu")
    else:
        vocoder = None

    is_multi_speaker = matcha.n_spks > 1

    dummy_input, input_names = get_inputs(is_multi_speaker)
    model, output_names = get_exportable_module(matcha, vocoder, args.n_timesteps)

    # Set dynamic shape for inputs/outputs
    dynamic_axes = {
        "x": {0: "batch_size", 1: "time"},
        "x_lengths": {0: "batch_size"},
    }

    if vocoder is None:
        dynamic_axes.update(
            {
                "mel": {0: "batch_size", 2: "time"},
                "mel_lengths": {0: "batch_size"},
            }
        )
    else:
        print("Embedding the vocoder in the ONNX graph")
        dynamic_axes.update(
            {
                "wav": {0: "batch_size", 1: "time"},
                "wav_lengths": {0: "batch_size"},
            }
        )

    if is_multi_speaker:
        dynamic_axes["spks"] = {0: "batch_size"}

    # Create the output directory (if not exists)
    Path(args.output).parent.mkdir(parents=True, exist_ok=True)

    model.to_onnx(
        args.output,
        dummy_input,
        input_names=input_names,
        output_names=output_names,
        dynamic_axes=dynamic_axes,
        opset_version=args.opset,
        export_params=True,
        do_constant_folding=True,
    )
    print(f"[🍵] ONNX model exported to  {args.output}")


if __name__ == "__main__":
    main()



import logging

import os
import unittest

from src.common.connector.multiprocessing_pipe import MultiprocessingPipeConnector
from src.common.logger import Logger

r"""
python -m unittest test.common.connector.test_mp_pipe
"""


class TestMultiprocessingPipeConnector(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        Logger.init(os.getenv("LOG_LEVEL", "debug").upper(), is_file=False)

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        self.connector = MultiprocessingPipeConnector()

    def tearDown(self):
        self.connector.close()

    def test_send_recv(self):
        print("be send")
        self.connector.send((1, 2, 3), "be")
        print("fe recv")
        v1, v2, v3 = self.connector.recv("fe")
        print(v1, v2, v3)
        self.assertEqual(v1, 1)
        self.assertEqual(v2, 2)
        self.assertEqual(v3, 3)



import os
import logging
import asyncio

import unittest

from src.common.logger import Logger
from src.common.session import Session
from src.common.types import LOG_DIR, SessionCtx
from src.common.connector.redis_queue import RedisQueueConnector


from dotenv import load_dotenv

load_dotenv(override=True)


r"""
REDIS_PASSWORD=*** python -m unittest test.common.connector.test_redis_queue
"""


class TestRedisQueue(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.REDIS_HOST = os.getenv(
            "REDIS_HOST", "redis-12259.c240.us-east-1-3.ec2.redns.redis-cloud.com"
        )
        cls.REDIS_PORT = os.getenv("REDIS_PORT", "12259")
        Logger.init(os.getenv("LOG_LEVEL", "debug").upper(), is_file=False)

    @classmethod
    def tearDownClass(cls):
        pass

    def setUp(self):
        self.connector = RedisQueueConnector(
            send_key="TEST_SEND", host=self.REDIS_HOST, port=self.REDIS_PORT
        )

    def tearDown(self):
        self.connector.close()

    def test_send_recv(self):
        self.connector.send((1, 2, 3), "be")
        v1, v2, v3 = self.connector.recv("fe")
        print(v1, v2, v3)
        self.assertEqual(v1, 1)
        self.assertEqual(v2, 2)
        self.assertEqual(v3, 3)

    def test_send_recv_session(self):
        s = Session(**SessionCtx("test_sid").__dict__)
        s.chat_history.append("hello")
        self.connector.send((1, 2, s), "be")
        v1, v2, session = self.connector.recv("fe")
        print(v1, v2, session)
        print(session.chat_history)
        self.assertEqual(v1, 1)
        self.assertEqual(v2, 2)
        self.assertIsNotNone(session)

        session.chat_history.append("world")
        self.connector.send((1, 2, session), "fe")
        v1, v2, s = self.connector.recv("be")
        print(v1, v2, s)
        print(s.chat_history)
        self.assertEqual(v1, 1)
        self.assertEqual(v2, 2)
        self.assertIsNotNone(s)



import torch
from device_cuda import CUDAInfo
from transformers import pipeline


def pipe_whisper_transcribe(audio_path, model_size="base", target_lang="zh"):
    info = CUDAInfo()

    # Initialize the ASR pipeline
    if info.is_cuda:
        pipe = pipeline(
            "automatic-speech-recognition",
            model=model_size,
            device="cuda:0",
            torch_dtype=torch.float16 if info.compute_capability_major >= 7 else torch.float32,
            model_kwargs={"use_flash_attention_2": info.compute_capability_major >= 8},
        )

        if info.compute_capability_major == 7 or info.compute_capability_major == 6:
            pipe.model = pipe.model.to_bettertransformer()
    else:
        pipe = pipeline(
            "automatic-speech-recognition",
            model=model_size,
            device="cpu",
            torch_dtype=torch.float32,
        )

        # for Word-level timestamps batch-size must be 1.
        # https://huggingface.co/openai/whisper-large-v3/discussions/12
        outputs = pipe(
            audio_path,
            chunk_length_s=30,
            batch_size=1,
            generate_kwargs={
                "language": target_lang,
            },
            return_timestamps="word",
        )
        print(outputs)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--model_type",
        "-t",
        type=str,
        default="whisper",
        help="choice whisper | whisper_timestamped",
    )
    parser.add_argument(
        "--audio_path", "-a", type=str, default="./records/tmp.wav", help="audio path"
    )
    parser.add_argument(
        "--model_path_or_size",
        "-m",
        type=str,
        default="./models/openai/whisper-base",
        help="model path or size",
    )
    parser.add_argument("--lang", "-l", type=str, default="zh", help="target language")
    args = parser.parse_args()
    pipe_whisper_transcribe(args.audio_path, args.model_path_or_size, args.lang)



import requests
from PIL import Image
import io
# Endpoint URLs
understand_image_url = "http://localhost:8000/understand_image_and_question/"
generate_images_url = "http://localhost:8000/generate_images/"

# Use your image file path here
image_path = "images/equation.png"

# Function to call the image understanding endpoint
def understand_image_and_question(image_path, question, seed=42, top_p=0.95, temperature=0.1):
    files = {'file': open(image_path, 'rb')}
    data = {
        'question': question,
        'seed': seed,
        'top_p': top_p,
        'temperature': temperature
    }
    response = requests.post(understand_image_url, files=files, data=data)
    response_data = response.json()
    print("Image Understanding Response:", response_data['response'])


# Function to call the text-to-image generation endpoint
def generate_images(prompt, seed=None, guidance=5.0):
    data = {
        'prompt': prompt,
        'seed': seed,
        'guidance': guidance
    }
    response = requests.post(generate_images_url, data=data, stream=True)
    
    if response.ok:
        img_idx = 1

        # We will create a new BytesIO for each image
        buffers = {}

        try:
            for chunk in response.iter_content(chunk_size=1024):
                if chunk:
                    # Use a boundary detection to determine new image start
                    if img_idx not in buffers:
                        buffers[img_idx] = io.BytesIO()

                    buffers[img_idx].write(chunk)

                    # Attempt to open the image
                    try:
                        buffer = buffers[img_idx]
                        buffer.seek(0)
                        image = Image.open(buffer)
                        img_path = f"generated_image_{img_idx}.png"
                        image.save(img_path)
                        print(f"Saved: {img_path}")

                        # Prepare the next image buffer
                        buffer.close()
                        img_idx += 1

                    except Exception as e:
                        # Continue loading data into the current buffer
                        continue

        except Exception as e:
            print("Error processing image:", e)
    else:
        print("Failed to generate images.")


# Example usage
if __name__ == "__main__":
    # Call the image understanding API
    understand_image_and_question(image_path, "What is this image about?")
    
    # Call the image generation API
    generate_images("A beautiful sunset over a mountain range, digital art.")



import re
import logging
import os
import unicodedata
from typing import List

import typer
import pymupdf
from rich.console import Console

from .table import table


app = typer.Typer()


class PDFExtractor:
    def extract_content(self, file_path: str) -> str:
        try:
            doc = pymupdf.open(file_path)
            content = " ".join(page.get_text() for page in doc)
            doc.close()

            # Normalize the text to handle special characters and remove accents
            normalized_content = unicodedata.normalize("NFKD", content)

            return normalized_content
        except Exception as e:
            logging.error(f"Error extracting PDF content: {str(e)}")
            raise


def get_pdf_file_name(file: str):
    pattern = r"([^/]+\.pdf)$"
    match = re.search(pattern, file)
    if match:
        pdf_file_name = match.group(1)
        return pdf_file_name
    else:
        raise Exception("must use *.pdf file")


@app.command()
def extract_content(
    pdf_files: List[str],
    output_dir: str = "videos/transcripts/",
) -> None:
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    extractor = PDFExtractor()
    for pdf_file in pdf_files:
        try:
            pdf_name = get_pdf_file_name(pdf_file)
            content = extractor.extract_content(pdf_file)
            print(f"PDF {pdf_file} content extracted successfully:")
            print(content[:500] + "..." if len(content) > 500 else content)
            # Save transcript to file
            output_file = os.path.join(output_dir, f"{pdf_name}.txt")
            with open(output_file, "w") as file:
                file.write(content)
        except Exception as e:
            print(f"An error occurred: {str(e)}")


@app.command()
def instruct_content(test_urls: List[str], language: str = "en") -> None:
    console = Console()
    extractor = PDFExtractor()
    for url in test_urls:
        try:
            with console.status("[bold green]Processing URL...") as status:
                content = extractor.extract_content(url)
                status.update("[bold blue]Generating Clips...")
                clips = table.extract_models(content, language=language)
                table.console_table(clips)

            console.print("\nChapter extraction complete!")
        except Exception as e:
            logging.error(f"An error occurred while processing {url}: {str(e)}")


r"""
python -m demo.content_parser.pdf_extractor_instructor extract-content \
    "/Users/wuyong/Documents/论文/llm/Attention Is All You Need.pdf" \
    "/Users/wuyong/Documents/论文/llm/《深度学习入门：基于Python的理论与实现》高清中文版.pdf"

python -m demo.content_parser.pdf_extractor_instructor extract-content \
    "/Users/wuyong/Desktop/iOS_18_All_New_Features_Sept_2024.pdf"

python -m demo.content_parser.pdf_extractor_instructor instruct-content \
    "/Users/wuyong/Documents/论文/llm/Attention Is All You Need.pdf" \
    "/Users/wuyong/Documents/论文/llm/《深度学习入门：基于Python的理论与实现》高清中文版.pdf" \
    --language zh

python -m demo.content_parser.pdf_extractor_instructor instruct-content \
    "/Users/wuyong/Desktop/iOS_18_All_New_Features_Sept_2024.pdf" \
    --language zh

#TODO: use OCR model extract PDF content e.g.: OCR-GOT2.0 :)
"""
if __name__ == "__main__":
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s - %(name)s - %(levelname)s - %(pathname)s:%(lineno)d - %(funcName)s - %(message)s",
        handlers=[
            # logging.FileHandler("extractor.log"),
            logging.StreamHandler()
        ],
    )
    app()



import json
import logging
import re
from typing import Dict, Tuple, List, Literal, Callable, Optional
import sys

from numba import jit
import numpy as np

from .utils import del_all


@jit
def _find_index(table: np.ndarray, val: np.uint16):
    for i in range(table.size):
        if table[i] == val:
            return i
    return -1


@jit
def _fast_replace(
    table: np.ndarray, text: bytes
) -> Tuple[np.ndarray, List[Tuple[str, str]]]:
    result = np.frombuffer(text, dtype=np.uint16).copy()
    replaced_words = []
    for i in range(result.size):
        ch = result[i]
        p = _find_index(table[0], ch)
        if p >= 0:
            repl_char = table[1][p]
            result[i] = repl_char
            replaced_words.append((chr(ch), chr(repl_char)))
    return result, replaced_words


class Normalizer:
    def __init__(self, map_file_path: str, logger=logging.getLogger(__name__)):
        self.logger = logger
        self.normalizers: Dict[str, Callable[[str], str]] = {}
        self.homophones_map = self._load_homophones_map(map_file_path)
        """
        homophones_map

        Replace the mispronounced characters with correctly pronounced ones.

        Creation process of homophones_map.json:

        1. Establish a word corpus using the [Tencent AI Lab Embedding Corpora v0.2.0 large] with 12 million entries. After cleaning, approximately 1.8 million entries remain. Use ChatTTS to infer the text.
        2. Record discrepancies between the inferred and input text, identifying about 180,000 misread words.
        3. Create a pinyin to common characters mapping using correctly read characters by ChatTTS.
        4. For each discrepancy, extract the correct pinyin using [python-pinyin] and find homophones with the correct pronunciation from the mapping.

        Thanks to:
        [Tencent AI Lab Embedding Corpora for Chinese and English Words and Phrases](https://ai.tencent.com/ailab/nlp/en/embedding.html)
        [python-pinyin](https://github.com/mozillazg/python-pinyin)

        """
        self.coding = "utf-16-le" if sys.byteorder == "little" else "utf-16-be"
        self.reject_pattern = re.compile(r"[^\u4e00-\u9fffA-Za-z，。、,\. ]")
        self.sub_pattern = re.compile(r"\[uv_break\]|\[laugh\]|\[lbreak\]")
        self.chinese_char_pattern = re.compile(r"[\u4e00-\u9fff]")
        self.english_word_pattern = re.compile(r"\b[A-Za-z]+\b")
        self.character_simplifier = str.maketrans(
            {
                "：": "，",
                "；": "，",
                "！": "。",
                "（": "，",
                "）": "，",
                "【": "，",
                "】": "，",
                "『": "，",
                "』": "，",
                "「": "，",
                "」": "，",
                "《": "，",
                "》": "，",
                "－": "，",
                ":": ",",
                ";": ",",
                "!": ".",
                "(": ",",
                ")": ",",
                "[": ",",
                "]": ",",
                ">": ",",
                "<": ",",
                "-": ",",
            }
        )
        self.halfwidth_2_fullwidth = str.maketrans(
            {
                "!": "！",
                '"': "“",
                "'": "‘",
                "#": "＃",
                "$": "＄",
                "%": "％",
                "&": "＆",
                "(": "（",
                ")": "）",
                ",": "，",
                "-": "－",
                "*": "＊",
                "+": "＋",
                ".": "。",
                "/": "／",
                ":": "：",
                ";": "；",
                "<": "＜",
                "=": "＝",
                ">": "＞",
                "?": "？",
                "@": "＠",
                # '[': '［',
                "\\": "＼",
                # ']': '］',
                "^": "＾",
                # '_': '＿',
                "`": "｀",
                "{": "｛",
                "|": "｜",
                "}": "｝",
                "~": "～",
            }
        )

    def __call__(
        self,
        text: str,
        do_text_normalization=True,
        do_homophone_replacement=True,
        lang: Optional[Literal["zh", "en"]] = None,
    ) -> str:
        if do_text_normalization:
            _lang = self._detect_language(text) if lang is None else lang
            if _lang in self.normalizers:
                text = self.normalizers[_lang](text)
            if _lang == "zh":
                text = self._apply_half2full_map(text)
        invalid_characters = self._count_invalid_characters(text)
        if len(invalid_characters):
            self.logger.warning(f"found invalid characters: {invalid_characters}")
            text = self._apply_character_map(text)
        if do_homophone_replacement:
            arr, replaced_words = _fast_replace(
                self.homophones_map,
                text.encode(self.coding),
            )
            if replaced_words:
                text = arr.tobytes().decode(self.coding)
                repl_res = ", ".join([f"{_[0]}->{_[1]}" for _ in replaced_words])
                self.logger.info(f"replace homophones: {repl_res}")
        if len(invalid_characters):
            text = self.reject_pattern.sub("", text)
        return text

    def register(self, name: str, normalizer: Callable[[str], str]) -> bool:
        if name in self.normalizers:
            self.logger.warning(f"name {name} has been registered")
            return False
        try:
            val = normalizer("test string 测试字符串")
            if not isinstance(val, str):
                self.logger.warning("normalizer must have caller type (str) -> str")
                return False
        except Exception as e:
            self.logger.warning(e)
            return False
        self.normalizers[name] = normalizer
        return True

    def unregister(self, name: str):
        if name in self.normalizers:
            del self.normalizers[name]

    def destroy(self):
        del_all(self.normalizers)
        del self.homophones_map

    def _load_homophones_map(self, map_file_path: str) -> np.ndarray:
        with open(map_file_path, "r", encoding="utf-8") as f:
            homophones_map: Dict[str, str] = json.load(f)
        map = np.empty((2, len(homophones_map)), dtype=np.uint32)
        for i, k in enumerate(homophones_map.keys()):
            map[:, i] = (ord(k), ord(homophones_map[k]))
        del homophones_map
        return map

    def _count_invalid_characters(self, s: str):
        s = self.sub_pattern.sub("", s)
        non_alphabetic_chinese_chars = self.reject_pattern.findall(s)
        return set(non_alphabetic_chinese_chars)

    def _apply_half2full_map(self, text: str) -> str:
        return text.translate(self.halfwidth_2_fullwidth)

    def _apply_character_map(self, text: str) -> str:
        return text.translate(self.character_simplifier)

    def _detect_language(self, sentence: str) -> Literal["zh", "en"]:
        chinese_chars = self.chinese_char_pattern.findall(sentence)
        english_words = self.english_word_pattern.findall(sentence)

        if len(chinese_chars) > len(english_words):
            return "zh"
        else:
            return "en"



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

"""Torch modules."""

# flake8: noqa
from .conv import (
    pad1d,
    unpad1d,
    NormConv1d,
    NormConvTranspose1d,
    NormConv2d,
    NormConvTranspose2d,
    SConv1d,
    SConvTranspose1d,
)
from .lstm import SLSTM
from .seanet import SEANetEncoder, SEANetDecoder
from .transformer import StreamingTransformerEncoder



# -*- coding: utf-8 -*-
"""DIGIT类
数字串 <=> 中文字符串 方法
中文字符串 <=> 数字串 方法
"""

__author__ = "Zhiyang Zhou <zyzhou@stu.xmu.edu.cn>"
__data__ = "2019-05-03"

from fish_speech.text.chn_text_norm.basic_util import *


class Digit:
    """
    DIGIT类
    """

    def __init__(self, digit=None, chntext=None):
        self.digit = digit
        self.chntext = chntext

    # def chntext2digit(self):
    #     return chn2num(self.chntext)

    def digit2chntext(self):
        return num2chn(self.digit, alt_two=False, use_units=False)


if __name__ == "__main__":

    # 测试程序
    print(Digit(digit="2016").digit2chntext())



import io
import os
import copy
import json
import logging
import torch
import transformers
import boto3
from typing import List, Optional, Tuple, Union, Dict, Sequence
from torch.utils.data import Dataset
from PIL import Image, ImageFile
ImageFile.LOAD_TRUNCATED_IMAGES = True
from GOT.utils.constants import *



class BaseDataset(Dataset):
    def __init__(
        self, 
        datasets: str,
        tokenizer: transformers.PreTrainedTokenizer,
        multimodal_cfg: dict
    ):
        super(BaseDataset, self).__init__()
        self.tokenizer = tokenizer
        self.multimodal_cfg = multimodal_cfg

        logging.warning(f"Using {multimodal_cfg['image_token_len']} tokens for representing image")

    def image_processor(self, image):
        # processor = self.multimodal_cfg['image_processor']  # the first processor, usually is the clip pretrained model (vit)
        processor_high = self.multimodal_cfg['image_processor_high'] # the second processor, usually is the designed image encoder (sam/swin/cnn)
        image_high = image.copy()

        #  Vary old codes
        
        # # TODO the 'keep', 'padding' only used for the first processor
        # if self.multimodal_cfg['image_aspect_ratio'] == 'keep':
        #     max_hw, min_hw = max(image.size), min(image.size)
        #     aspect_ratio = max_hw / min_hw
        #     max_len, min_len = 448, 224
        #     shortest_edge = int(min(max_len / aspect_ratio, min_len))
        #     image = processor.preprocess(image, return_tensors='pt', do_center_crop=False, size={"shortest_edge": shortest_edge})['pixel_values'][0]
        # elif self.multimodal_cfg['image_aspect_ratio'] == 'pad':
        #     def expand2square(pil_img, background_color):
        #         width, height = pil_img.size
        #         if width == height:
        #             return pil_img
        #         elif width > height:
        #             result = Image.new(pil_img.mode, (width, width), background_color)
        #             result.paste(pil_img) # for simpler box processing
        #             return result
        #         else:
        #             result = Image.new(pil_img.mode, (height, height), background_color)
        #             result.paste(pil_img) # for simpler box processing
        #             return result
        #     image = expand2square(image, tuple(int(x*255) for x in processor.image_mean))
        #     image = processor.preprocess(image, return_tensors='pt', do_center_crop=False, size={"shortest_edge": 224})['pixel_values'][0]
        # else:
        #     image = processor.preprocess(image, return_tensors='pt')['pixel_values'][0]

        image_high = processor_high(image_high)

        return image_high

    def __len__(self):
        return len(self.list_data_dict)

    def __getitem__(self, i) -> Dict[str, torch.Tensor]:
        pass


import logging

import torch

try:
    from transformers import MimiModel, AutoFeatureExtractor
except ModuleNotFoundError as e:
    logging.error(
        "In order to use transoformers mimi codec, you need to `pip install achatbot[codec_transformers_mimi]`."
    )
    raise Exception(f"Missing module: {e}")

from achatbot.common.factory import EngineClass
from achatbot.common.utils.helper import get_device, print_model_params
from achatbot.types.codec import CodecArgs
from achatbot.modules.codec.interface import ICodec


class TransformersMimiCodec(EngineClass, ICodec):
    """
    use transoformers[torch] mimi model and config
    https://huggingface.co/kyutai/mimi/blob/main/config.json
    """

    TAG = "codec_transformers_mimi"

    def __init__(self, **kwargs) -> None:
        super().__init__()
        self.args = CodecArgs(**kwargs)
        self.args.device = self.args.device or get_device()
        logging.info("CodecArgs: %s", self.args)
        self.load_model()

    def load_model(self):
        self.model = MimiModel.from_pretrained(self.args.model_dir)
        self.model.eval().to(self.args.device)
        print_model_params(self.model, "mimi")
        self.feature_extractor = AutoFeatureExtractor.from_pretrained(self.args.model_dir)
        logging.debug(f"feature_extractor: {self.feature_extractor}")

    @torch.no_grad
    def encode_code(self, waveform_tensor: torch.Tensor) -> torch.Tensor:
        return self.model.encode(waveform_tensor[None, None, :].to(self.args.device)).audio_codes
        # pre-process the input waveform
        inputs = self.feature_extractor(
            raw_audio=waveform_tensor,
            sampling_rate=self.feature_extractor.sampling_rate,
            return_tensors="pt",
        )

        output = self.model.encode(inputs["input_values"].to(self.args.device))
        vq_codes = output.audio_codes
        logging.debug(f"encode waveform to vq_codes: {vq_codes.shape}")
        return vq_codes

    @torch.no_grad
    def decode_code(self, vq_codes: torch.Tensor) -> torch.Tensor:
        waveform_tensor = self.model.decode(vq_codes.to(self.args.device))[0]
        logging.debug(f"decode vq_codes to gen waveform: {waveform_tensor.shape}")
        return waveform_tensor[0][0]



import logging
import os
from typing import List


try:
    import torch
    from snac import SNAC
except ImportError as e:
    logging.error("In order to use SNAC-codec, you need to `pip install achatbot[codec_snac]`.")
    raise ImportError(f"Missing module: {e}")

from src.common.factory import EngineClass
from src.common.utils.helper import get_device, print_model_params
from src.types.codec import CodecArgs
from src.modules.codec.interface import ICodec


class SNACCodec(EngineClass, ICodec):
    """
    Multi-Scale Neural Audio Codec (SNAC) compresses audio into discrete codes at a low bitrate
    - https://arxiv.org/abs/2410.14411
    - https://github.com/hubertsiuzdak/snac
    - https://huggingface.co/hubertsiuzdak/snac_24khz (for 🗣️ Speech)
    - https://huggingface.co/hubertsiuzdak/snac_32khz (for 🎸 Music / Sound Effects)
    - https://huggingface.co/hubertsiuzdak/snac_44khz (for 🎸 Music / Sound Effects)
    """

    TAG = "codec_snac"

    def __init__(self, **kwargs) -> None:
        super().__init__()
        self.args = CodecArgs(**kwargs)
        self.args.device = self.args.device or get_device()
        logging.info("CodecArgs: %s", self.args)
        self.load_model()

    def load_model(self):
        self.model = SNAC.from_pretrained(self.args.model_dir).to(self.args.device).eval()
        print_model_params(self.model, "snac")

    @torch.inference_mode()
    def encode_code(self, waveform_tensor: torch.Tensor) -> List[torch.Tensor]:
        vq_codes_list = self.model.encode(
            waveform_tensor[None, None, :].to(self.args.device)
        )  # waveform_tensor: [1,1,T]
        logging.debug(f"encode waveform to vq_codes cn: {len(vq_codes_list)}")
        return vq_codes_list

    @torch.inference_mode()
    def decode_code(self, vq_codes: List[torch.Tensor]) -> torch.Tensor:
        if not isinstance(vq_codes, list):
            vq_codes = vq_codes.to(self.args.device)
            vq_codes = [vq_codes.squeeze(0)]  # [[1,T]]
        waveform_tensor = self.model.decode(vq_codes)
        logging.debug(f"decode vq_codes to gen waveform: {waveform_tensor.shape}")
        return waveform_tensor[0][0]



# https://pytorch.org/blog/training-moes/
import os
import modal


app = modal.App("train-demo-ep")

demo_image = modal.Image.debian_slim(python_version="3.12").pip_install("torch")


@app.function(
    image=demo_image,
    gpu=os.getenv("IMAGE_GPU", "T4:2"),
    retries=0,
    # how long should we stay up with no requests?
    scaledown_window=15 * 60,
)
def run(num_experts=4, top_k=2, alpha=0.01, mode="brodcast"):
    import torch

    world_size = torch.cuda.device_count()  # 使用所有可用 GPU
    if world_size < 2:
        print("需要至少 2 个 GPU 来演示分布式数据并行")
        return
    if num_experts < top_k:
        print(f"top_k {top_k} 必须 小于 总专家数 {num_experts}")
        return

    print(
        f"Running Expert Parallelism mode {mode} with {world_size} GPUs, Experts/GPU: {num_experts//world_size}, top_k: {top_k}, load balance loss alpha: {alpha}"
    )

    train_ep(world_size, num_experts, top_k, alpha, mode)


"""
# defualt mode=brodcast num-experts=2, top-k=1, alpha=0.0 on 2 GPU
modal run src/train/demo/ep.py
modal run src/train/demo/ep.py --num-experts 2 --top-k 2 --alpha 0.1
modal run src/train/demo/ep.py --num-experts 2 --top-k 1 --alpha 0.1
modal run src/train/demo/ep.py --num-experts 4 --top-k 2 --alpha 0.01

modal run src/train/demo/ep.py --mode all2all
modal run src/train/demo/ep.py --mode all2all --num-experts 2 --top-k 2 --alpha 0.1
modal run src/train/demo/ep.py --mode all2all --num-experts 2 --top-k 1 --alpha 0.1
modal run src/train/demo/ep.py --mode all2all --num-experts 4 --top-k 2 --alpha 0.01
"""


@app.local_entrypoint()
def main(num_experts: str = "2", top_k: str = "1", alpha: str = "0.0", mode="brodcast"):
    run.remote(int(num_experts), int(top_k), float(alpha), mode)


# 训练函数
def train_ep(world_size, num_experts=4, top_k=2, alpha=0.01, mode="brodcast"):
    import os
    from multiprocessing import Process
    import torch
    import torch.nn as nn
    import torch.distributed as dist

    # 初始化分布式环境
    def setup_distributed(rank, world_size):
        os.environ["MASTER_ADDR"] = "localhost"
        os.environ["MASTER_PORT"] = "12355"
        dist.init_process_group("nccl", rank=rank, world_size=world_size)
        torch.cuda.set_device(rank)

    # 清理分布式环境
    def cleanup():
        dist.destroy_process_group()

    # 定义单个专家（一个简单的线性层）
    class Expert(nn.Module):
        def __init__(self, input_size, hidden_size):
            super(Expert, self).__init__()
            self.fc = nn.Linear(input_size, hidden_size)
            self.relu = nn.ReLU()

        def forward(self, x):
            return self.relu(self.fc(x))

    # 定义门控网络
    class GatingNetwork(nn.Module):
        def __init__(self, input_size, num_experts):
            super(GatingNetwork, self).__init__()
            self.fc = nn.Linear(input_size, num_experts)  # 输出每个专家的得分

        def forward(self, x):
            return torch.softmax(self.fc(x), dim=-1)  # 返回每个专家的概率

    # MoE 模型
    class MoE(nn.Module):
        def __init__(
            self, input_size, hidden_size, num_experts, rank, world_size, top_k, alpha=0.01
        ):
            super(MoE, self).__init__()
            self.rank = rank
            self.world_size = world_size
            self.num_experts = num_experts
            self.top_k = top_k
            self.alpha = alpha  # 负载均衡损失的权重
            self.experts_per_gpu = num_experts // world_size
            self.device = torch.device(f"cuda:{rank}")

            # 门控网络在所有设备上运行
            self.gating = GatingNetwork(input_size, num_experts)

            # 每个 GPU 分配部分专家
            self.local_experts = nn.ModuleList(
                [Expert(input_size, hidden_size) for _ in range(self.experts_per_gpu)]
            ).to(self.device)

            # 本地专家的全局索引范围
            self.start_expert_idx = rank * self.experts_per_gpu
            self.end_expert_idx = (rank + 1) * self.experts_per_gpu
            print(f"rank {rank} expert_idx range [{self.start_expert_idx} {self.end_expert_idx})")

        def compute_load_balance_loss(self, gate_scores, topk_indices):
            # 计算每个专家的使用频率（fraction of tokens）
            batch_size = gate_scores.size(0)
            expert_usage = torch.zeros(self.num_experts, device=gate_scores.device)
            for i in range(batch_size):
                for idx in topk_indices[i]:
                    expert_usage[idx] += 1
            f = expert_usage / (batch_size * self.top_k)  # 归一化使用频率

            # 计算门控得分的平均值（gate probability）
            p = gate_scores.mean(dim=0)  # [num_experts]

            # 负载均衡损失：f 和 p 的乘积方差
            load_loss = torch.mean(f * p) ** 2 + torch.var(f * p)
            return self.alpha * load_loss

        def forward(self, x):
            # 门控网络计算每个专家的得分
            gate_scores = self.gating(x)  # [batch_size, num_experts]

            # 选择 Top-k 专家
            topk_scores, topk_indices = torch.topk(
                gate_scores, self.top_k, dim=-1
            )  # [batch_size, k]

            # 计算负载均衡损失
            load_balance_loss = self.compute_load_balance_loss(gate_scores, topk_indices)
            if self.rank == 0:
                pass
                # print once
                # print(f"topk_scores {topk_scores}, topk_indices {topk_indices}")
                # print(f"rank {self.rank} load_balance_loss {load_balance_loss}")

            # 按目标 GPU 分组输入
            inputs_to_send = [[] for _ in range(self.world_size)]
            output_buffers = [[] for _ in range(self.world_size)]
            local_outputs = []

            batch_size = x.size(0)
            for i in range(batch_size):
                for j in range(self.top_k):
                    expert_idx = topk_indices[i, j].item()
                    target_rank = expert_idx // self.experts_per_gpu
                    score = topk_scores[i, j]
                    if target_rank == self.rank:
                        # 本地专家处理
                        local_expert_idx = expert_idx - self.start_expert_idx
                        local_output = self.local_experts[local_expert_idx](
                            x[i : i + 1]
                        )  # 处理单个样本
                        local_outputs.append((i, j, local_output.squeeze(0) * score))
                    else:
                        # 发送到其他 GPU
                        inputs_to_send[target_rank].append((i, j, x[i]))
                        # print(f"rank {self.rank} send to {target_rank} {i} {x[i]} {j}")

            # print(
            #    f"rank {self.rank} local_outputs {local_outputs} \ninputs_to_send {inputs_to_send}"
            # )

            # 同步接收输入
            recv_buffers = []
            for r in range(self.world_size):
                if r != self.rank and inputs_to_send[r]:
                    x_tensor = torch.stack([x for _, _, x in inputs_to_send[r]])
                    ctx_expert_idx = torch.tensor(
                        [[i, j] for i, j, _ in inputs_to_send[r]],
                        dtype=torch.long,
                        device=self.device,
                    )  # 对应的专家索引
                    send_tensor = torch.cat((ctx_expert_idx, x_tensor), dim=1)

                    num_inputs = torch.tensor(
                        [len(inputs_to_send[r])], dtype=torch.long, device=self.device
                    )
                    dist.broadcast(
                        num_inputs, src=self.rank
                    )  # 将对应的样本大小发给对应GPU的专家处理
                    dist.broadcast(send_tensor, src=self.rank)  # 将对应的样本发给对应GPU的专家处理
                    # print(f"rank {self.rank} send to {r} -> {send_tensor.shape}")
                if r != self.rank:
                    recv_num_inputs = torch.zeros(1, dtype=torch.long, device=self.device)
                    dist.broadcast(recv_num_inputs, src=r)
                    # print(f"rank {self.rank} recv num_inputs {recv_num_inputs} from {r}")
                    if recv_num_inputs.item() > 0:
                        recv_tensor = torch.zeros(
                            recv_num_inputs.item(), 2 + x.size(1), device=self.device
                        )
                        # 接收对应的样本,给对应GPU的专家处理
                        dist.broadcast(recv_tensor, src=r)
                        # print(f"rank {self.rank} recv recv_tensor {recv_tensor} <- from {r}")
                        recv_buffers.append((r, recv_tensor))

            # 处理接收到的样本
            for src_rank, recv_tensor in recv_buffers:
                for j in range(recv_tensor.size(0)):
                    ctx_expert_idx_i = int(recv_tensor[j : j + 1][0, 0].item())
                    ctx_expert_idx_j = int(recv_tensor[j : j + 1][0, 1].item())
                    expert_idx = topk_indices[ctx_expert_idx_i, ctx_expert_idx_j].item()
                    local_expert_idx = expert_idx - self.start_expert_idx
                    if local_expert_idx >= self.end_expert_idx:
                        print(
                            f"[WARNING] rank {self.rank} local_expert_idx {local_expert_idx} >= {self.end_expert_idx},don't belong to this rank"
                        )
                        continue
                    output = self.local_experts[local_expert_idx](
                        recv_tensor[j : j + 1][:, 2:]
                    )  # 处理单个样本 [1,input_size] -> [1,hidden_size]
                    output_buffers[src_rank].append(
                        (ctx_expert_idx_i, ctx_expert_idx_j, output.squeeze(0))
                    )

            # dist.barrier()

            # 发送输出回原 GPU
            for r in range(self.world_size):
                if r != self.rank and output_buffers[r]:
                    x_tensor = torch.stack([x for _, _, x in output_buffers[r]])
                    ctx_expert_idx = torch.tensor(
                        [[i, j] for i, j, _ in output_buffers[r]],
                        dtype=torch.long,
                        device=self.device,
                    )  # 对应的专家索引
                    send_tensor = torch.cat((ctx_expert_idx, x_tensor), dim=1)
                    dist.broadcast(send_tensor, src=self.rank)
                    # print(f"rank {self.rank} send to dst {r} {send_tensor.shape}")

            # 接收返回的输出
            final_topk_output = torch.zeros(
                batch_size, self.top_k, self.local_experts[0].fc.out_features, device=self.device
            )  # [batch_size, top_k, hidden_size]
            for r in range(self.world_size):
                if r != self.rank and inputs_to_send[r]:
                    num_outputs = len(inputs_to_send[r])
                    recv_tensor = torch.zeros(
                        num_outputs, 2 + final_topk_output.size(2), device=self.device
                    )
                    dist.broadcast(recv_tensor, src=r)
                    # print(
                    #    f"rank {self.rank} final_topk_output: {final_topk_output.shape} inputs_to_send: {len(inputs_to_send[r])}, recv_tensor: {recv_tensor.shape}"
                    # )
                    for j in range(recv_tensor.size(0)):
                        ctx_expert_idx_i = int(recv_tensor[j : j + 1][0, 0].item())
                        ctx_expert_idx_j = int(recv_tensor[j : j + 1][0, 1].item())
                        score = topk_scores[ctx_expert_idx_i, ctx_expert_idx_j]
                        out = recv_tensor[j : j + 1][:, 2:]
                        final_topk_output[ctx_expert_idx_i][ctx_expert_idx_j] += (
                            out.squeeze(0) * score
                        )  # Top-k 加权累加

            # 合并本地输出
            for i, j, out in local_outputs:
                final_topk_output[i][j] += out  # Top-k 加权累加

            # print(final_topk_output)
            # [batch_size, top_k, hidden_size] -> [batch_size, hidden_size]
            final_output = torch.mean(final_topk_output, dim=1)
            # final_output = torch.sum(final_topk_output, dim=1)
            # print(final_output)

            return final_output, load_balance_loss

    class All2AllMoE(MoE):
        def forward(self, x):
            batch_size = x.size(0)
            gate_scores = self.gating(x)  # [batch_size, num_experts]
            topk_scores, topk_indices = torch.topk(
                gate_scores, self.top_k, dim=-1
            )  # [batch_size, k]

            # 计算负载均衡损失
            load_balance_loss = self.compute_load_balance_loss(gate_scores, topk_indices)

            # 统计每个 GPU 需要处理的样本数
            counts_per_rank = torch.zeros(self.world_size, dtype=torch.long, device=self.device)
            for i in range(batch_size):
                for j in range(self.top_k):
                    expert_idx = topk_indices[i, j].item()
                    target_rank = expert_idx // self.experts_per_gpu
                    counts_per_rank[target_rank] += 1

            # 准备发送的输入张量和元数据
            send_tensors = [[] for _ in range(self.world_size)]
            send_indices = [[] for _ in range(self.world_size)]
            send_scores = [[] for _ in range(self.world_size)]
            for i in range(batch_size):
                for j in range(self.top_k):
                    expert_idx = topk_indices[i, j].item()
                    target_rank = expert_idx // self.experts_per_gpu
                    send_tensors[target_rank].append(x[i])
                    send_indices[target_rank].append(i)
                    send_scores[target_rank].append(topk_scores[i, j])

            # 将列表转换为张量列表，并填充到最大长度
            max_count = counts_per_rank.max().item()
            send_list = [
                torch.stack(send_tensors[r])
                if len(send_tensors[r]) > 0
                else torch.zeros(max_count, x.size(1), device=self.device)
                for r in range(self.world_size)
            ]
            recv_list = [
                torch.zeros(max_count, x.size(1), device=self.device)
                for _ in range(self.world_size)
            ]

            # 使用 all_to_all 分发输入
            dist.all_to_all(recv_list, send_list)

            # 处理接收到的输入
            local_outputs = torch.zeros(
                batch_size, self.local_experts[0].fc.out_features, device=self.device
            )
            for i in range(counts_per_rank[self.rank]):
                if i < recv_list[self.rank].size(0):
                    expert_idx = topk_indices[
                        send_indices[self.rank][i], i // counts_per_rank[self.rank] % self.top_k
                    ].item()
                    if self.start_expert_idx <= expert_idx < self.end_expert_idx:
                        local_expert_idx = expert_idx - self.start_expert_idx
                        output = self.local_experts[local_expert_idx](
                            recv_list[self.rank][i : i + 1]
                        )
                        orig_idx = send_indices[self.rank][i]
                        score = send_scores[self.rank][i]
                        local_outputs[orig_idx] += output.squeeze(0) * score

            # 使用 all_to_all 收集输出
            send_output_list = [
                torch.zeros(max_count, local_outputs.size(1), device=self.device)
                for _ in range(self.world_size)
            ]
            for i in range(counts_per_rank[self.rank]):
                if i < len(send_indices[self.rank]):
                    send_output_list[self.rank][i] = local_outputs[send_indices[self.rank][i]]

            recv_output_list = [
                torch.zeros(max_count, local_outputs.size(1), device=self.device)
                for _ in range(self.world_size)
            ]
            dist.all_to_all(recv_output_list, send_output_list)

            # 聚合最终输出
            final_output = torch.zeros(batch_size, local_outputs.size(1), device=self.device)
            for r in range(self.world_size):
                for i in range(counts_per_rank[r]):
                    if i < len(send_indices[r]):
                        orig_idx = send_indices[r][i]
                        score = send_scores[r][i]
                        final_output[orig_idx] += recv_output_list[r][i] * score
            return final_output, load_balance_loss

    # 专家并行训练函数
    def train_expert_parallel(
        rank,
        world_size,
        num_experts,
        input_size,
        hidden_size,
        data,
        labels,
        top_k,
        alpha=0.01,
        mode="brodcast",
    ):
        # 设置分布式环境
        setup_distributed(rank, world_size)

        # 参数设置
        device = torch.device(f"cuda:{rank}")

        # 将数据移动到设备
        data, labels = data.to(device), labels.to(device)

        # 初始化 MoE 模型
        if mode == "all2all":
            model = All2AllMoE(
                input_size, hidden_size, num_experts, rank, world_size, top_k, alpha
            ).to(device)
        else:
            model = MoE(input_size, hidden_size, num_experts, rank, world_size, top_k, alpha).to(
                device
            )

        # 定义损失函数和优化器
        criterion = nn.MSELoss()  # 假设回归任务
        # criterion = nn.CrossEntropyLoss()  # 分类任务
        optimizer = torch.optim.SGD(model.parameters(), lr=0.01)

        # 训练步骤
        model.train()
        for epoch in range(5):  # 模拟 5 个 epoch
            optimizer.zero_grad()

            # 前向传播
            output, load_balance_loss = model(data)
            task_loss = criterion(output, labels)

            # 总损失 = 任务损失 + 负载均衡损失
            loss = task_loss + load_balance_loss

            # 反向传播
            loss.backward()
            optimizer.step()

            # 打印损失
            dist.barrier()  # 同步所有进程
            if rank == 0:
                print(
                    f"Epoch {epoch+1}, Loss: {loss.item():.4f} Task Loss: {task_loss.item():.4f} Load Balance Loss: {load_balance_loss.item():.4f}"
                )

        cleanup()

    # 生成模拟数据
    batch_size = 4
    input_size = 2
    hidden_size = 6
    data = torch.randn(batch_size, input_size, requires_grad=True)
    labels = torch.randn(batch_size, hidden_size, requires_grad=True)  # 启用梯度

    # 启动进程
    processes = []
    for rank in range(world_size):
        p = Process(
            target=train_expert_parallel,
            args=(
                rank,
                world_size,
                num_experts,
                input_size,
                hidden_size,
                data,
                labels,
                top_k,
                alpha,
                mode,
            ),
        )
        p.start()
        processes.append(p)

    for p in processes:
        p.join()



import torch


class TorchSeedContext:
    def __init__(self, seed):
        self.seed = seed
        self.state = None

    def __enter__(self):
        self.state = torch.random.get_rng_state()
        torch.manual_seed(self.seed)

    def __exit__(self, type, value, traceback):
        torch.random.set_rng_state(self.state)



# Copyright (c) 2024 Alibaba Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from torch.utils.data import DataLoader
from inspiremusic.dataset.dataset import Dataset
import numpy as np
import librosa

def audio_process_dataset_and_dataloader(args, configs):
    input_dataset = Dataset(args.input_data, data_pipeline=configs['data_pipeline'], mode='processing', shuffle=True, partition=True)
    # do not use persistent_workers=True, as whisper tokenizer opens tiktoken file each time when the for loop starts
    input_data_loader = DataLoader(input_dataset,
                                   batch_size=None,
                                   pin_memory=args.pin_memory,
                                   num_workers=args.num_workers,
                                   prefetch_factor=args.prefetch)
    return input_dataset, input_data_loader

def is_silent(wav_path, threshold=0.01, frame_length=2048, hop_length=512):
    y, sr = librosa.load(wav_path, sr=None)
    rms = librosa.feature.rms(y=y, frame_length=frame_length, hop_length=hop_length)[0]
    silent_frames = np.sum(rms < threshold) / len(rms)
    silence_fraction_threshold = 0.95
    return silent_frames >= silence_fraction_threshold

def rich_captions(text=None, tags=None, lyrics=None, chorus="verse", start_time=0.0, end_time=30.0):
    if text is None and tags is None and lyrics is None:
        return None
    else:
        if start_time is None:
            start_time = 0.0
        if end_time is None:
            end_time = 30.0
        if chorus is None:
            chorus = "verse"
        captions = f"<|{start_time:.1f}|><|{chorus}|>"
        if tags is not None:
            captions += f"<|{tags}|>"
        if text is not None:
            captions += f"<|{text}|>"
        if lyrics is not None:
            captions += f"<|lyrics|><|{lyrics}|>"
        captions += f"<|{end_time:.1f}|>"
        return captions

def process_tags(infile, outfile, timefile = None):
    key_list = []
    with open(infile, "r") as f:
        for line in f:
            sec = line.strip()
            key_list.append(sec)
    f.close()
    if timefile is None:
        with open(outfile, 'w') as f:
            for k in key_list:
                parts = k.rsplit('_', 1)
                text = parts[0].replace('_', ' ') + ', ' + parts[1]
                caption = rich_captions(text, None, None)
                if caption is not None:
                    f.write("%s\t%s\n" %(k, caption))
        f.close()        
    else:
        times = {}
        with open(timefile, "r") as f:
            for line in f:
                sec = line.strip().split("\t")
                if len(sec) == 2 :
                    times[sec[0]] = sec[1]
        f.close()

        with open(outfile, 'w') as f:
            for k in key_list:
                parts = k.rsplit('_', 1)
                text = parts[0].replace('_', ' ') + ', ' + parts[1]
                if k in times.keys():
                    caption = rich_captions(text, None, None, "verse", 0.0, float(times[k]))
                    if caption is not None:
                        f.write("%s\t%s\n" %(k, caption))
        f.close()

def process_trans(infile, outfile):
    trans = {}
    with open(infile, "r") as f:
        for line in f:
            sec = line.strip().split("\t")
            if len(sec) == 2:
                trans[sec[0]] = sec[1]
            else:
                print(line)
    f.close()
    with open(outfile, 'w') as f:
        for k, v in trans.items():
            f.write("%s\t%s\n" %(k, rich_captions(v)))
    f.close()